#!/usr/bin/env python
"""
wx widgets for Larch
"""
__DOC__ = '''
WxPython functions for larch

function         description
------------     ------------------------------
gcd              graphical change directory - launch browser to select working folder
fileprompt       launch file browser to select files.

'''

import platform
import os

HAS_WXPYTHON = False
try:
    import wx
    HAS_WXPYTHON = True
except (ImportError, AttributeError):
    HAS_WXPYTHON = False

_larch_name = '_sys.wx'
_larch_builtins = {}

FONTSIZE = 9
if platform.system() == 'Windows':
    FONTSIZE = 10
if platform.system() == 'Darwin':
    FONTSIZE = 12

if HAS_WXPYTHON:

    from . import larchframe
    from . import larchfilling
    from . import readlinetextctrl

    from wxutils import (set_sizer, pack, SetTip, Font, HLine, Check, MenuItem,
                         Popup, is_wxPhoenix, RIGHT, LEFT, CEN ,
                         LTEXT, FRAMESTYLE, hms, DateTimeCtrl, Button,
                         ToggleButton, BitmapButton, Choice, YesNo, SimpleText,
                         TextCtrl, LabeledTextCtrl, HyperText, EditableListBox,
                         get_icon, GUIColors, OkCancel, FileOpen, FileSave,
                         SelectWorkdir, fix_filename, SavedParameterDialog,
                         FileCheckList, FileDropTarget, GridPanel, RowPanel,
                         make_steps, set_float, FloatCtrl)

    from .floats import NumericCombo, FloatSpin, FloatSpinWithPin

    from .notebooks import flatnotebook

    from .larchframe import LarchFrame, LarchPanel
    from .columnframe import ColumnDataFileFrame, EditColumnFrame
    from .athena_importer import AthenaImporter
    from .reportframe import ReportFrame
    from .gui_utils import gcd, databrowser, fileprompt, wx_update

    from .parameter import ParameterWidgets, ParameterPanel
    from .periodictable import PeriodicTablePanel

    _larch_builtins = {'_sys.wx': dict(gcd=gcd,
                                       databrowser=databrowser,
                                       filepromspt=fileprompt,
                                       wx_update=wx_update)}

    from .plotter import (_plot, _oplot, _newplot, _plot_text,
                          _plot_marker, _plot_arrow, _plot_setlimits,
                          _plot_axvline, _plot_axhline, _scatterplot,
                          _hist, _update_trace, _saveplot, _saveimg,
                          _getDisplay, _closeDisplays, _getcursor,
                          last_cursor_pos, _imshow, _contour, _xrf_plot,
                          _xrf_oplot, _fitplot, _redraw_plot)

    from .xrfdisplay import  XRFDisplayFrame

    from . import xafsplots
    from .xafsplots import plotlabels

    _larch_builtins['_plotter'] = dict(plot=_plot, oplot=_oplot,
                                       newplot=_newplot, plot_text=_plot_text,
                                       plot_marker=_plot_marker,
                                       plot_arrow=_plot_arrow,
                                       plot_setlimits=_plot_setlimits,
                                       plot_axvline=_plot_axvline,
                                       plot_axhline=_plot_axhline,
                                       scatterplot=_scatterplot, hist=_hist,
                                       update_trace=_update_trace,
                                       save_plot=_saveplot,
                                       save_image=_saveimg,
                                       get_display=_getDisplay,
                                       close_all_displays=_closeDisplays,
                                       get_cursor=_getcursor,
                                       last_cursor_pos=last_cursor_pos,
                                       imshow=_imshow, contour=_contour,
                                       xrf_plot=_xrf_plot,
                                       xrf_oplot=_xrf_oplot,
                                       fit_plot=_fitplot,
                                       redraw_plot=_redraw_plot)

    _larch_builtins['_xafs'] = dict(redraw=xafsplots.redraw,
                                    plot_mu=xafsplots.plot_mu,
                                    plot_bkg=xafsplots.plot_bkg,
                                    plot_chie=xafsplots.plot_chie,
                                    plot_chik=xafsplots.plot_chik,
                                    plot_chir=xafsplots.plot_chir,
                                    plot_chifit=xafsplots.plot_chifit,
                                    plot_path_k=xafsplots.plot_path_k,
                                    plot_path_r=xafsplots.plot_path_r,
                                    plot_paths_k=xafsplots.plot_paths_k,
                                    plot_paths_r=xafsplots.plot_paths_r,
                                    plot_diffkk=xafsplots.plot_diffkk,
                                    plot_prepeaks_fit=xafsplots.plot_prepeaks_fit,
                                    plot_prepeaks_baseline=xafsplots.plot_prepeaks_baseline,
                                    plot_pca_components=xafsplots.plot_pca_components,
                                    plot_pca_weights=xafsplots.plot_pca_weights,
                                    plot_pca_fit=xafsplots.plot_pca_fit)


    def _larch_init(_larch):
        """add ScanFrameViewer to _sys.gui_apps """
        if _larch is None:
            return
        _sys = _larch.symtable._sys
        if not hasattr(_sys, 'gui_apps'):
            _sys.gui_apps = {}
        _sys.gui_apps['xrfviewer'] = ('XRF Spectrum Viewer', XRFDisplayFrame)


    #############################
    ## Hack System and Startfile on Windows totry to track down
    ## weird error of starting other applications, like Mail
    if platform.system() == 'Windows':
        from os import system as os_system
        from os import startfile as os_startfile
        def my_system(command):
            print("#@ os.system: ", command)
            return os_system(command)

        def my_startfile(filepath, operation=None):
            print("#@ os.startfile: ", filepath, operation)
            try:
                if operation is None:
                    return os_startfile(filepath)
                else:
                    return os_startfile(filepath, operation)
            except WindowsError:
                print("#@ os.startfile failed: ", filepath, operation)

        os.system = my_system
        os.startfile = my_startfile
    #############################

else:
    def nullfunc(*args, **kws):
        pass

    _larch_builtins = {'_sys.wx': dict(gcd=nullfunc,
                                       databrowser=nullfunc,
                                       filepromspt=nullfunc,
                                       wx_update=nullfunc)}

    _larch_builtins['_plotter'] = dict(plot=nullfunc,
                                       oplot=nullfunc,
                                       newplot=nullfunc,
                                       plot_text=nullfunc,
                                       plot_marker=nullfunc,
                                       plot_arrow=nullfunc,
                                       plot_setlimits=nullfunc,
                                       plot_axvline=nullfunc,
                                       plot_axhline=nullfunc,
                                       scatterplot=nullfunc,
                                       hist=nullfunc,
                                       update_trace=nullfunc,
                                       save_plot=nullfunc,
                                       save_image=nullfunc,
                                       get_display=nullfunc,
                                       close_all_displays=nullfunc,
                                       get_cursor=nullfunc,
                                       last_cursor_pos=nullfunc,
                                       imshow=nullfunc,
                                       contour=nullfunc,
                                       xrf_plot=nullfunc,
                                       xrf_oplot=nullfunc,
                                       fit_plot=nullfunc)



#!/usr/bin/env python
'''
Diffraction functions require for fitting and analyzing data.

mkak 2017.02.06 (originally written spring 2016)
'''


##########################################################################
# IMPORT PYTHON PACKAGES

import math

import numpy as np
from scipy import optimize,signal,interpolate

from .xrd_tools import (d_from_q, d_from_twth, twth_from_d, twth_from_q,
                        q_from_d, q_from_twth)


##########################################################################
# FUNCTIONS

def peakfilter(intthrsh,ipeaks,y,verbose=False):
    '''
    Returns x and y for data set corresponding to peak indices solution
    from peakfilter() with the option of setting a minimum intensity
    threshold for filtering peaks
    '''

    ipks = []
    ipks += [i for i in ipeaks if y[i] > intthrsh]
    if verbose: print('Peaks under intensity %i filtered out.' % intthrsh)

    return ipks

def peaklocater(ipeaks,x):
    '''
    Returns x and y for data set corresponding to peak indices solution
    from peakfinder()
    '''
    xypeaks = [x[i] for i in ipeaks]

    return np.array(xypeaks)

def peakfinder_methods():

    methods = []
    try:
        import peakutils
        methods += ['peakutils.indexes']
    except:
        pass
    try:
        from scipy import signal
        methods += ['scipy.signal.find_peaks_cwt']
    except:
        pass

    return methods


def peakfinder(y, method='scipy.signal.find_peaks_cwt',
               widths=20, gapthrsh=5, thres=0.0, min_dist=10):
    '''
    Returns indices for peaks in y from dataset
    '''

    if method == 'peakutils.indexes':
        try:
            import peakutils
        except:
            print('python package peakutils not installed')
            widths = np.arange(1,int(len(y)/widths))
            peak_indices = signal.find_peaks_cwt(y, widths, gap_thresh=gapthrsh)
        peak_indices = peakutils.indexes(y, thres=thres, min_dist=min_dist)
    elif method == 'scipy.signal.find_peaks_cwt':
        ## scipy.signal.find_peaks_cwt(vector, widths, wavelet=None, max_distances=None,
        ##                   gap_thresh=None, min_length=None, min_snr=1, noise_perc=10)
        widths = np.arange(1,int(len(y)/widths))
        peak_indices = signal.find_peaks_cwt(y, widths, gap_thresh=gapthrsh)

    return peak_indices

def peakfitter(ipeaks, twth, I, verbose=True, halfwidth=40, fittype='single'):

    peaktwth,peakFWHM,peakinty = [],[],[]
    for j in ipeaks:
        if j > halfwidth and (len(twth)-j) > halfwidth:
            minval = int(j - halfwidth)
            maxval = int(j + halfwidth)

            if I[j] > I[minval] and I[j] > I[maxval]:
                xdata = twth[minval:maxval]
                ydata = I[minval:maxval]

                try:
                    pktwth,pkfwhm,pkint = data_gaussian_fit(xdata,ydata,fittype=fittype)
                    peaktwth += [pktwth]
                    peakFWHM += [pkfwhm]
                    peakinty += [pkint]
                except:
                    pass

    return np.array(peaktwth),np.array(peakFWHM),np.array(peakinty)


def data_gaussian_fit(x,y,fittype='single'):
    '''
    Fits a single or double Gaussian functions.
    '''
    meanx = sum(x)/float(len(x))
    meany = sum(y)/float(len(y))
    sigma = np.sqrt(sum(y*(x-meanx)**2)/float(len(x)))

    try:
        popt,pcov = optimize.curve_fit(gaussian,x,y,p0=[np.max(y),meanx,sigma])
    except:
        popt = [1,1,1]

    if fittype == 'double':
        a,b,c = popt
        popt2,pcov2 = optimize.curve_fit(doublegaussian,x,y,p0=[a,b,c,np.min(y),meanx,sigma])

    rsqu_n1 = 0
    rsqu_n2 = 0
    rsqu_d = 0
    for i in range(x.shape[0]):
        rsqu_n1 = (y[i] - gaussian(x[i],*popt))**2 + rsqu_n1
        if fittype == 'double':
            rsqu_n2 = (y[i] - doublegaussian(x[i],*popt2))**2 + rsqu_n2
        rsqu_d = (y[i] - meany)**2 + rsqu_d


    if fittype == 'double':
        pkpos = popt2[1]
        pkfwhm = abs(2*np.sqrt(2*math.log1p(2))*popt2[2])
        pkint  = np.max(doublegaussian(x,*popt2))
    else:
        pkpos = popt[1]
        pkfwhm = abs(2*np.sqrt(2*math.log1p(2))*popt[2])
        pkint  = np.max(gaussian(x,*popt))

    return pkpos,pkfwhm,pkint



def gaussian(x,a,b,c):
    return a*np.exp(-(x-b)**2/(2*c**2))



def doublegaussian(x,a1,b1,c1,a2,b2,c2):
    return a1*np.exp(-(x-b1)**2/(2*c1**2))+a2*np.exp(-(x-b2)**2/(2*c2**2))


def instrumental_fit_uvw(ipeaks, twthaxis, I, halfwidth=40, verbose=True):

    twth,FWHM,inten = peakfitter(ipeaks,twthaxis,I,halfwidth=halfwidth,
                           fittype='double',verbose=verbose)

    tanth = np.tan(np.radians(twth/2))
    sqFWHM  = FWHM**2

    (u,v,w) = data_poly_fit(tanth,sqFWHM,verbose=verbose)

    if verbose:
        print('\nFit results:')
        for i,(twthi,fwhmi,inteni) in enumerate(zip(twth,FWHM,inten)):
            print('Peak %i @ %0.2f deg. (fwhm %0.3f deg, %i counts)' % (i,twthi,fwhmi,inteni))
    print('\nInstrumental broadening parameters:')
    print('---  U  : %0.8f'   % u)
    print('---  V  : %0.8f'   % v)
    print('---  W  : %0.8f\n' % w)

    return(u,v,w)


def poly_func(x,a,b,c):
    return a*x**2 + b*x + c


def data_poly_fit(x, y, plot=False, verbose=False):
    '''
    Fits a set of data with to a second order polynomial function.
    '''

    try:
        popt,pcov = optimize.curve_fit(poly_func,x,y,p0=[1,1,1])
    except:
        print( 'WARNING: scipy.optimize.curve_fit was unsuccessful.' )
        return [1,1,1]

    n = len(x)
    meany = sum(y)/n

    rsqu_n = 0
    rsqu_d = 0
    for i in range(x.shape[0]):
        rsqu_n = (y[i] - poly_func(x[i],*popt))**2 + rsqu_n
        rsqu_d = (y[i] - meany)**2 + rsqu_d

    if verbose:
        print('---Polynomial Fit:')
        print('---  U  : %0.8f'   % popt[0])
        print('---  V  : %0.8f'   % popt[1])
        print('---  W  : %0.8f\n' % popt[2])

    return popt

def trim_range(data,min,max,axis=0):

    maxi = -1
    mini = 0
    for i,val in enumerate(data[axis]):
        mini = i if val < min else mini
        maxi = i if val < max else maxi

    return data[:,mini:maxi]

def interpolate_for_y(x,x0,y0):
    t = interpolate.splrep(x0,y0)
    return interpolate.splev(x,t,der=0)

def calc_peak():

    x = np.arange(-2,2,0.001)

    ## Lorentzian
    b = 1/(1+(x**2))

    ## Gaussian
    c = np.exp(-math.log1p(2)*(x**2))

    ## VOIGT: Convolve, shift, and scale
    d = np.convolve(b,c,'same')
    d = norm_pattern(d,c)
    shiftx = find_max(x,d)
    newx = x-shiftx

    ## Diffraction pattern data.
    ## x,b    - 'Lorentzian'
    ## x,c    - 'Gaussian'
    ## newx,d - 'Voigt'

    return 'Lorentzian',b,x,'Gaussian',x,c,'Voigt',newx,d


def norm_pattern(intensity,scale_int):

    max_int = np.max(intensity)
    scale = np.max(scale_int)
    intensity = intensity/max_int*scale

    return(intensity)


def scale_100(intensity):

    intensity = norm_pattern(intensity,100)

    return(intensity)



def find_max(x,y):
    return [xi for xi,yi in zip(x,y) if yi == np.max(y)][0]



def calcRsqu(y,ycalc):

    ss_res = 0
    ss_tot = 0

    ymean = np.mean(y) #sum(y)/float(len(y))

    for i,yi in enumerate(y):
        ss_res = ss_res + (yi - ycalc[i])**2
        ss_tot = ss_tot + (yi - ymean)**2

    return (1 - (ss_res/ss_tot))


def outliers(y,scale=0.2):
    for i,yi in enumerate(y):
        if i > 0 and i < (len(y)-1):
#             if yi > y[i-1]*scale and yi > y[i+1]*scale:
#                 y[i] = (y[i-1]+y[i+1])/2
            if abs(yi-y[i-1]) > yi/scale and abs(yi - y[i+1]) > yi/scale:
                y[i] = (y[i-1]+y[i+1])/2

    return y


def calc_broadening(pklist, twth, wavelength, u=1.0, v=1.0, w=1.0, C=1.0, D=None, verbose=False,smooth=False):
    '''
    == Broadening calculation performed in 2theta - not q ==

    pklist     - location of peaks in range [q,2th,d,I]
    twth       -  2-theta axis
    wavelength - in units A

    u,v,w - instrumental broadening parameters
    C - shape factor (1 for sphere)
    D - particle size in units A (if D is None, no size broadening)
    '''

    ## TERMS FOR INSTRUMENTAL BROADENING
    thrad = np.radians(pklist[1]/2)
    Bi = np.sqrt( u*(np.tan(thrad))**2 + v*np.tan(thrad) + w )

    ## TERMS FOR SIZE BROADENING
    ## FWHM(2th) = (C*wavelength)/(D*math.cos(math.radians(twth/2)))
    ## FWHM(q)   = (C*wavelength)/D*(1/np.sqrt(1-termB))
    if D is not None:
        termB = ((wavelength*pklist[0])/(2*math.pi))**2
        Bs = (C*wavelength)/D*(1/np.sqrt(1-termB))

    ## Define intensity array for broadened peaks
    Itot = np.zeros(len(twth))

    step = twth[1]-twth[0]

    ## Loop through all peaks
    for i,peak in enumerate(zip(*pklist)):
        if peak[1] > min(twth) and peak[1] < max(twth):
            A = peak[3]
            B = peak[1]

            c_i = Bi[i]/abs(2*np.sqrt(2*math.log1p(2)))

            if D is not None: c_s = Bs[i]/abs(2*np.sqrt(2*math.log1p(2)))

            #    Bm = np.sqrt(Bi[i]**2+Bs[i]**2)
            #else:
            #    Bm = np.sqrt(Bi[i]**2)

            #c_m = Bm/abs(2*np.sqrt(2*math.log1p(2)))

            width1 = max(Bs[i],Bi[i]) if D is not None else Bi[i]
            width2 = min(Bs[i],Bi[i]) if D is not None else Bi[i]

            min2th = B-2*width1
            max2th = B+2*width1

            num = abs((max2th-min2th)/step)
            twthG = np.linspace(max2th,min2th,num)

            intBi = A*np.exp(-(twthG-B)**2/(2*c_i**2))
            if D is not None: intBs = A*np.exp(-(twthG-B)**2/(2*c_s**2))
            #intBm = A*np.exp(-(twthG-B)**2/(2*c_m**2))

            if D is not None:
                new_intensity = np.convolve(intBs,intBi,'same')
            else:
                new_intensity = intBi

            new_intensity = norm_pattern(new_intensity,A)

            X = twthG
            Y = new_intensity

            bins = len(twth)
            idx  = np.digitize(X,twth)
            Ii    = [np.sum(Y[idx==k]) for k in range(bins)]
            Itot = Itot + Ii

    if smooth:
        Itot = outliers(Itot)
    return Itot

##########################################################################
# def fit_with_minimization(q,I,parameters=None,fit_method='leastsq'):
#     '''
#     fit_method options: 'leastsq','cobyla','slsqp','nelder'
#
#     parameters of type Parameters(): needs a,b,c,nsize,pkshift?
#
#     my_pars = Parameters()
#     my_pars.add('nsize', value= NPsize, min= 3.0,  max=100.0,vary=False)
#     my_pars.add('a', value=bkgdA, min=minA, max=maxA)
#     my_pars.add('b', value=bkgdB, min=minB, max=maxB)
#     my_pars.add('c', value=bkgdC, min=minC, max=maxC)
#     '''
#
#     from lmfit import minimize,Parameters,report_fit
#
# #     ## First, fit background alone:
# #     result = minimize(BACKGROUND_FUNCTION, my_pars, args=(q,I),method=fit_method)
# #     ## Second, add fitted parameters into parameter set
# #     my_pars.add('?',  value=result.params['?'].value)
# #     ## Third, fit peaks on top of background
# #     result = minimize(BACKGROUND_FUNCTION+PEAKS_FUNCTION, my_pars, args=(q,I),method=fit_method)
# #     ## Fourth, view error report
# #     report_fit(result)
# #     ## Fifth, return results
# #     NPsize = result.params['nsize'].value
# #     bkgdA  = result.params['a'].value
# #     bkgdB  = result.params['b'].value
# #     bkgdC  = result.params['c'].value
#
#
#     return



#!/usr/bin/env python
"""
Epics XRF Display App
"""

import sys
import os

import time
import copy
from functools import partial

import wx
import wx.lib.mixins.inspection
import wx.lib.scrolledpanel as scrolled
try:
    from wx._core import PyDeadObjectError
except:
    PyDeadObjectError = Exception

import wx.lib.colourselect  as csel
import numpy as np
import matplotlib

HAS_PLOT = False
try:
    from wxmplot import PlotPanel
    HAS_PLOT = True
except ImportError:
    HAS_PLOT = False

HAS_DV = False
try:
    import wx.dataview as dv
    HAS_DV = True
except:
    pass

from wxutils import (SimpleText, EditableListBox, Font, FloatCtrl,
                     pack, Popup, Button, get_icon, Check, MenuItem,
                     Choice, FileOpen, FileSave, fix_filename, HLine,
                     GridPanel, CEN, LEFT, RIGHT)


import larch
from larch.site_config import icondir
from larch.wxlib import PeriodicTablePanel
from larch.wxlib.xrfdisplay import (XRFDisplayFrame, XRFCalibrationFrame,
                                    FILE_WILDCARDS)

ROI_WILDCARD = 'Data files (*.dat)|*.dat|ROI files (*.roi)|*.roi|All files (*.*)|*.*'
try:
    from epics import caget
    from .xrf_detectors import Epics_MultiXMAP, Epics_Xspress3
except:
    pass

HAS_SCANDB = False
try:
    from epicsscan import ScanDB
    HAS_SCANDB = True
except:
    pass

class DetectorSelectDialog(wx.Dialog):
    """Connect to an Epics MCA detector
    Can be either XIA xMAP  or Quantum XSPress3
    """
    msg = '''Select XIA xMAP or Quantum XSPress3 MultiElement MCA detector'''
    det_types = ('SXD-7', 'ME-7', 'ME-4', 'other')
    ioc_types = ('Xspress3.1', 'xMAP', 'Xspress3.0')
    def_prefix = '13QX7:'   # SDD1:'
    def_nelem  =  4

    def __init__(self, parent=None, prefix=None, det_type='ME-4',
                 ioc_type='Xspress3', nmca=4,
                 title='Select Epics MCA Detector'):
        if prefix is None:
            prefix = self.def_prefix
        if det_type not in self.det_types:
            det_type = self.det_types[0]

        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title)

        self.SetBackgroundColour((240, 240, 230))
        self.SetFont(Font(9))
        if parent is not None:
            self.SetFont(parent.GetFont())

        self.ioctype = Choice(self,size=(120, -1), choices=self.ioc_types)
        self.ioctype.SetStringSelection(ioc_type)

        self.dettype = Choice(self,size=(120, -1), choices=self.det_types)
        self.dettype.SetStringSelection(det_type)

        self.prefix = wx.TextCtrl(self, -1, prefix, size=(120, -1))
        self.nelem = FloatCtrl(self, value=nmca, precision=0, minval=1,
                               size=(120, -1))

        btnsizer = wx.StdDialogButtonSizer()

        if wx.Platform != "__WXMSW__":
            btn = wx.ContextHelpButton(self)
            btnsizer.AddButton(btn)

        btn = wx.Button(self, wx.ID_OK)
        btn.SetHelpText("Use this detector")
        btn.SetDefault()
        btnsizer.AddButton(btn)

        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()

        hline = wx.StaticLine(self, size=(225, 3), style=wx.LI_HORIZONTAL)
        sty = LEFT
        sizer = wx.GridBagSizer(5, 2)
        def txt(label):
            return SimpleText(self, label, size=(120, -1), style=LEFT)

        sizer.Add(txt('  Detector Type'),  (0, 0), (1, 1), sty, 2)
        sizer.Add(txt('  Uses Xspress3?'), (1, 0), (1, 1), sty, 2)
        sizer.Add(txt('  Epics Prefix'),  (2, 0), (1, 1), sty, 2)
        sizer.Add(txt('  # Elements'),    (3, 0), (1, 1), sty, 2)
        sizer.Add(self.dettype,         (0, 1), (1, 1), sty, 2)
        sizer.Add(self.ioctype,         (1, 1), (1, 1), sty, 2)
        sizer.Add(self.prefix,          (2, 1), (1, 1), sty, 2)
        sizer.Add(self.nelem,           (3, 1), (1, 1), sty, 2)

        sizer.Add(hline,                (4, 0), (1, 2), sty, 2)
        sizer.Add(btnsizer,             (5, 0), (1, 2), sty, 2)
        self.SetSizer(sizer)
        sizer.Fit(self)


class EpicsXRFDisplayFrame(XRFDisplayFrame):
    _about = """Epics XRF Spectra Display
  Matt Newville <newville @ cars.uchicago.edu>
  """
    me4_layout = ((0, 0), (1, 0), (1, 1), (0, 1))
    main_title = 'Epics XRF Control'

    def __init__(self, parent=None, _larch=None, prefix=None,
                 det_type='ME-4', ioc_type='Xspress3', nmca=4,
                 size=(725, 580), environ_file=None, scandb_conn=None,
                 title='Epics XRF Display', output_title='XRF', **kws):

        self.det_type = det_type
        self.ioc_type = ioc_type
        self.nmca = nmca
        self.det_fore = 1
        self.det_back = 0
        self.scandb = None
        self.environ = []
        if environ_file is not None:
            self.read_environfile(environ_file)
        if HAS_SCANDB and scandb_conn is not None:
            self.ConnectScanDB(**scandb_conn)

        self.onConnectEpics(event=None, prefix=prefix)

        self.icon_file = os.path.join(icondir, 'ptable.ico')

        XRFDisplayFrame.__init__(self, parent=parent, _larch=_larch,
                                 title=title, size=size, **kws)

    def read_environfile(self, filename):
        """read environmnet file"""
        if os.path.exists(filename):
            textlines = []
            try:
                with open(filename, 'r') as fh:
                    textlines = fh.readlines()
            except IOError:
                return
            self.environ = []
            for line in textlines:
                line = line[:-1].replace('\t', ' ')
                pvname, desc = line.split(' ', 1)
                desc = desc.strip()
                self.environ.append((pvname, desc))

    def onConnectEpics(self, event=None, prefix=None, **kws):
        if prefix is None:
            res  = self.prompt_for_detector(prefix=prefix,
                                            ioc_type=self.ioc_type,
                                            nmca=self.nmca)
            self.prefix, self.det_type, self.ioc_type, self.nmca = res
        else:
            self.prefix = prefix
        self.det_fore = 1
        self.det_back = 0
        self.clear_mcas()
        self.connect_to_detector(prefix=self.prefix, ioc_type=self.ioc_type,
                                 det_type=self.det_type, nmca=self.nmca)

    def ConnectScanDB(self, **kws):
        if not HAS_SCANDB:
            return
        self.scandb = ScanDB(**kws)
        if self.scandb is not None:
            basedir = self.scandb.get_info('user_folder')
            fileroot = self.scandb.get_info('server_fileroot')
        basedir = str(basedir)
        fileroot = str(fileroot)
        if basedir.startswith(fileroot):
            basedir = basedir[len(fileroot):]
        fullpath = os.path.join(fileroot, basedir)
        fullpath = fullpath.replace('\\', '/').replace('//', '/')
        curdir = os.getcwd()
        try:
            os.chdir(fullpath)
        except:
            os.chdir(curdir)
        self.scandb.connect_pvs()

    def onSaveMCAFile(self, event=None, **kws):
        tmp = '''
        # print('SaveMCA File')
        deffile = ''
        if hasattr(self.mca, 'sourcefile'):
            deffile = "%s%s" % (deffile, getattr(self.mca, 'sourcefile'))
        if hasattr(self.mca, 'areaname'):
            deffile = "%s%s" % (deffile, getattr(self.mca, 'areaname'))
        if deffile == '':
            deffile ='test'
        if not deffile.endswith('.mca'):
            deffile = deffile + '.mca'
        '''

        deffile = 'save.mca' # fix_filename(str(deffile))
        outfile = FileSave(self, "Save MCA File",
                           default_file=deffile,
                           wildcard=FILE_WILDCARDS)

        environ = []
        if HAS_SCANDB and self.scandb is not None:
            c, table = self.scandb.get_table('pvs')
            pvrows = self.scandb.query(table).all()
            for row in pvrows:
                addr = str(row.name)
                desc = str(row.notes)
                val  = self.scandb.pvs[addr].get(as_string=True)
                environ.append((addr, val, desc))

        elif len(self.environ) > 0:
            for pvname, desc in self.environ:
                val  = caget(pvname, as_string=True)
                environ.append((pvname, val, desc))

        if outfile is not None:
            self.det.save_mcafile(outfile, environ=environ)

    def onSaveColumnFile(self, event=None, **kws):
        print( '  EPICS-XRFDisplay onSaveColumnFile not yet implemented  ')
        pass

    def prompt_for_detector(self, prefix=None, ioc_type='Xspress3',  nmca=4):
        dlg = DetectorSelectDialog(prefix=prefix, ioc_type=ioc_type, nmca=nmca)
        dlg.Raise()
        if dlg.ShowModal() == wx.ID_OK:
            dpref = dlg.prefix.GetValue()
            atype = dlg.ioctype.GetStringSelection()
            dtype = dlg.dettype.GetStringSelection()
            nmca = dlg.nelem.GetValue()
            dlg.Destroy()
        return dpref, dtype, atype, nmca

    def connect_to_detector(self, prefix=None, ioc_type='Xspress3',
                            det_type=None, nmca=4):
        self.det = None
        ioc_type = ioc_type.lower()
        if ioc_type.startswith('xspress3'):
            version = 2
            if 'old' in ioc_type:
                version = 1
            self.det = Epics_Xspress3(prefix=prefix, nmca=nmca, version=version)
            self.det.connect()
            time.sleep(0.5)
            self.det.get_mca(mca=1)
            self.needs_newplot=True
        else:
            self.det = Epics_MultiXMAP(prefix=prefix, nmca=nmca)

    def show_mca(self, init=False):
        self.needs_newplot = False
        if self.mca is None or self.needs_newplot:
            self.mca = self.det.get_mca(mca=self.det_fore)

        self.plotmca(self.mca, set_title=False, init=init)
        title = "Foreground: MCA{:d}".format(self.det_fore)
        if self.det_back  > 0:
            if self.mca2 is None:
                self.mca2 = self.det.get_mca(mca=self.det_back)

            c2 = self.det.get_array(mca=self.det_back)
            e2 = self.det.get_energy(mca=self.det_back)
            title = "{:s}  Background: MCA{:d}".format(title, self.det_back)
            try:
                self.oplot(e2, c2)
            except ValueError:
                pass

        roiname = self.get_roiname()

        if roiname in self.wids['roilist'].GetStrings():
            i = self.wids['roilist'].GetStrings().index(roiname)
            self.wids['roilist'].EnsureVisible(i)
            self.onROI(label=roiname)
        dtime = self.det.get_deadtime(mca=self.det_fore)
        if dtime is not None:
            self.wids['deadtime'].SetLabel("%.1f" % dtime)
        self.SetTitle("%s: %s" % (self.main_title, title))
        self.needs_newplot = False

    def onSaveROIs(self, event=None, **kws):
        dlg = wx.FileDialog(self, message="Save ROI File",
                            defaultDir=os.getcwd(),
                            wildcard=ROI_WILDCARD,
                            style = wx.FD_SAVE|wx.FD_CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            roifile = dlg.GetPath()

        self.det.save_rois(roifile)

    def onRestoreROIs(self, event=None, **kws):
        dlg = wx.FileDialog(self, message="Read ROI File",
                            defaultDir=os.getcwd(),
                            wildcard=ROI_WILDCARD,
                            style = wx.FD_OPEN|wx.FD_CHANGE_DIR)

        if dlg.ShowModal() == wx.ID_OK:
            roifile = dlg.GetPath()
            self.det.restore_rois(roifile)
            self.set_roilist(mca=self.mca)
            self.show_mca()
            self.onSelectDet(event=None, index=0)

    def createCustomMenus(self):
        menu = wx.Menu()
        MenuItem(self, menu, "Connect to Detector\tCtrl+D",
                 "Connect to MCA or XSPress3 Detector",
                 self.onConnectEpics)
        menu.AppendSeparator()
        self._menus.insert(1, (menu, 'Detector'))

    def createMainPanel(self):
        epicspanel = self.createEpicsPanel()
        ctrlpanel  = self.createControlPanel()
        plotpanel  = self.panel = self.createPlotPanel()
        self.panel.SetName('plotpanel')
        tx, ty = self.wids['ptable'].GetBestSize()
        cx, cy = ctrlpanel.GetBestSize()
        px, py = plotpanel.GetBestSize()

        self.SetSize((950, 625))
        self.SetMinSize((450, 350))

        style = wx.ALIGN_LEFT|wx.EXPAND|wx.ALL

        bsizer = wx.BoxSizer(wx.HORIZONTAL)
        bsizer.Add(ctrlpanel, 0, style, 1)
        bsizer.Add(plotpanel, 1, style, 1)
        hline = wx.StaticLine(self, size=(425, 2), style=wx.LI_HORIZONTAL|style)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(epicspanel, 0, style, 1)
        sizer.Add(hline,      0, style, 1)
        sizer.Add(bsizer,     1, style, 1)
        pack(self, sizer)

        try:
            self.SetIcon(wx.Icon(self.icon_file, wx.BITMAP_TYPE_ICO))
        except:
            pass
        self.set_roilist(mca=None)

    def create_detbuttons(self, pane):
        btnpanel = wx.Panel(pane, name='buttons')
        btnsizer = wx.GridBagSizer(1, 1)        
        btns = {}
        sx = 30
        sy = int(sx/2)
        for i in range(1, self.nmca+1):
            b = Button(btnpanel, '%d' % i, size=(sx, sx),
                       action=partial(self.onSelectDet, index=i))
            b.SetFont(Font(10))
            self.wids['det%i' % i] = b
            btns[i] = b
        dtype = self.det_type.lower().replace('-', '').replace(' ', '').replace('_', '')
           
        if dtype.startswith('sxd7') and self.nmca == 7:
            btnsizer.Add((sx, sy), (0, 0), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[4],  (1, 0), (2, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[5],  (3, 0), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 0), (1, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[3],  (0, 2), (2, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[7],  (2, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[6],  (4, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (0, 4), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[2],  (1, 4), (2, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[1],  (3, 4), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 4), (1, 2), wx.ALIGN_LEFT, 1)
        elif dtype.startswith('me7') and self.nmca == 7:
            btnsizer.Add((sx, sy), (0, 0), (1, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[7],  (1, 0), (2, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[6],  (3, 0), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 0), (1, 2), wx.ALIGN_LEFT, 1)                        
            btnsizer.Add(btns[2],  (0, 2), (2, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[1],  (2, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[5],  (4, 2), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (0, 4), (1, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add(btns[3],  (1, 4), (2, 2), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[4],  (3, 4), (2, 2), wx.ALIGN_LEFT, 1)
            btnsizer.Add((sx, sy), (5, 4), (1, 2), wx.ALIGN_LEFT, 1)
        elif dtype.startswith('me4') and self.nmca == 4:
            btnsizer.Add(btns[1],  (0, 0), (1, 1), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[2],  (1, 0), (1, 1), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[3],  (1, 1), (1, 1), wx.ALIGN_LEFT, 1)            
            btnsizer.Add(btns[4],  (0, 1), (1, 1), wx.ALIGN_LEFT, 1)
        else:
            NPERROW = 4
            icol, irow = 0, 0
            for nmca  in range(1, self.nmca+1):
                btnsizer.Add(btns[nmca],  (irow, icol), (1, 1), wx.ALIGN_LEFT, 1)            
                icol += 1
                if icol > NPERROW-1:
                    icol = 0
                    irow += 1

        pack(btnpanel, btnsizer)
        return btnpanel
        
    def createEpicsPanel(self):
        pane = wx.Panel(self, name='epics panel')
        style  = wx.ALIGN_LEFT
        rstyle = wx.ALIGN_RIGHT
       
        det_btnpanel = self.create_detbuttons(pane)

        bkg_choices = ['None'] + ["%d" % (i+1) for i in range(self.nmca)]

        self.wids['det_status'] = SimpleText(pane, ' ', size=(120, -1), style=style)
        self.wids['deadtime']   = SimpleText(pane, ' ', size=(120, -1), style=style)

        self.wids['bkg_det'] = Choice(pane, size=(100, -1), choices=bkg_choices,
                                      action=self.onSelectDet)

        self.wids['dwelltime'] = FloatCtrl(pane, value=0.0, precision=1, minval=0,
                                           size=(80, -1), act_on_losefocus=True,
                                           action=self.onSetDwelltime)
        self.wids['elapsed'] = SimpleText(pane, ' ', size=(80, -1),  style=style)

        self.wids['mca_sum'] = Choice(pane, size=(100, -1),
                                      choices=['Single', 'Accumulate'],
                                      action=self.onMcaSumChoice,
                                      default=1 )

        b1 =  Button(pane, 'Start',      size=(90, -1), action=self.onStart)
        b2 =  Button(pane, 'Stop',       size=(90, -1), action=self.onStop)
        b3 =  Button(pane, 'Erase',      size=(90, -1), action=self.onErase)
        b4 =  Button(pane, 'Continuous', size=(90, -1), action=partial(self.onStart,
                                                                       dtime=0.0))

        sum_lab = SimpleText(pane, 'Accumulate Mode:',   size=(150, -1))
        bkg_lab = SimpleText(pane, 'Background MCA:',   size=(150, -1))
        pre_lab = SimpleText(pane, 'Dwell Time (s):',   size=(125, -1))
        ela_lab = SimpleText(pane, 'Elapsed Time (s):', size=(125, -1))
        sta_lab = SimpleText(pane, 'Status :',          size=(100, -1))
        dea_lab = SimpleText(pane, '% Deadtime:',       size=(100, -1))

        psizer = wx.GridBagSizer(5, 5)
        psizer.Add(SimpleText(pane, ' MCAs: '),  (0, 0), (1, 1), style, 1)
        psizer.Add(det_btnpanel,           (0, 1), (2, 1), style, 1)
        psizer.Add(bkg_lab,                (0, 2), (1, 1), style, 1)
        psizer.Add(self.wids['bkg_det'],   (0, 3), (1, 1), style, 1)
        psizer.Add(sum_lab,                (1, 2), (1, 1), style, 1)
        psizer.Add(self.wids['mca_sum'],   (1, 3), (1, 1), style, 1)
        psizer.Add(pre_lab,                (0, 4), (1, 1),  style, 1)
        psizer.Add(ela_lab,                (1, 4), (1, 1),  style, 1)
        psizer.Add(self.wids['dwelltime'], (0, 5), (1, 1),  style, 1)
        psizer.Add(self.wids['elapsed'],   (1, 5), (1, 1),  style, 1)

        psizer.Add(b1, (0, 6), (1, 1), style, 1)
        psizer.Add(b4, (0, 7), (1, 1), style, 1)
        psizer.Add(b2, (1, 6), (1, 1), style, 1)
        psizer.Add(b3, (1, 7), (1, 1), style, 1)

        psizer.Add(sta_lab,                  (0, 8), (1, 1), style, 1)
        psizer.Add(self.wids['det_status'],  (0, 9), (1, 1), style, 1)
        psizer.Add(dea_lab,                  (1, 8), (1, 1), style, 1)
        psizer.Add(self.wids['deadtime'],    (1, 9), (1, 1), style, 1)
        pack(pane, psizer)
        # pane.SetMinSize((500, 53))
        self.det.connect_displays(status=self.wids['det_status'],
                                  elapsed=self.wids['elapsed'])

        wx.CallAfter(self.onSelectDet, index=1, init=True)
        self.timer_counter = 0
        self.mca_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.UpdateData, self.mca_timer)
        self.mca_timer.Start(100)
        return pane

    def UpdateData(self, event=None, force=False):
        self.timer_counter += 1
        if self.mca is None or self.needs_newplot:
            self.show_mca()
        # self.elapsed_real = self.det.elapsed_real
        self.mca.real_time = self.det.elapsed_real
        # print("Update Data  ", force, self.det.needs_refresh)

        if force or self.det.needs_refresh:
            self.det.needs_refresh = False
            if self.det_back > 0:
                if self.mca2 is None:
                    self.mca2 = self.det.get_mca(mca=self.det_back)

                counts = self.det.get_array(mca=self.det_back)
                energy = self.det.get_energy(mca=self.det_back)
                try:
                    self.update_mca(counts, energy=energy, is_mca2=True, draw=False)
                except ValueError:
                    pass

            if self.mca is None:
                self.mca = self.det.get_mca(mca=self.det_fore)

            dtime = self.det.get_deadtime(mca=self.det_fore)
            if dtime is not None:
                self.wids['deadtime'].SetLabel("%.1f" % dtime)

            counts = self.det.get_array(mca=self.det_fore)*1.0
            energy = self.det.get_energy(mca=self.det_fore)
            if max(counts) < 1.0:
                counts    = 1e-4*np.ones(len(counts))
                counts[0] = 2.0
            self.update_mca(counts, energy=energy)

    def ShowROIStatus(self, left, right, name='', panel=0):
        if left > right:
            return
        sum = self.ydata[left:right].sum()

        try:
            ftime, nframes = self.det.get_frametime()
        except:
            ftime   = self.det.frametime
            nframes = self.det.nframes
        self.det.elapsed_real = nframes * ftime

        mca_counts = self.det.mcas[self.det_fore-1].get('VAL')
        sum =  mca_counts[left:right].sum()
        # print("ROI STATUS ", name, ftime, nframes, sum, cps, mca_counts.sum(),  mca_counts)
        if name in (None, ''):
            name = 'Selected'
        else:
            for roi in self.det.mcas[self.det_fore-1].rois:
                if name.lower() == roi.name.lower():
                    try:
                        sum = roi.sum
                    except:
                        pass
        cps = sum/ftime
        if cps < 0: cps = 0
        # print("ROI STATUS ", name, _counts, cps)
        fmt = " {:s}: Cts={:10,.0f} :{:10,.1f} Hz"
        self.write_message(fmt.format(name, sum, cps), panel=panel)

    def onSelectDet(self, event=None, index=0, init=False, **kws):
        if index > 0:
            self.det_fore = index
        self.det_back = self.wids['bkg_det'].GetSelection()
        if self.det_fore  == self.det_back:
            self.det_back = 0

        for i in range(1, self.nmca+1):
            dname = 'det%i' % i
            bcol = (210, 210, 210)
            fcol = (0, 0, 0)
            if i == self.det_fore:
                fcol = (200,  20,  20)
                bcol = (250, 250, 250)
            self.wids[dname].SetBackgroundColour(bcol)
            self.wids[dname].SetForegroundColour(fcol)
        self.clear_mcas()
        self.show_mca(init=init)
        self.Refresh()

    def swap_mcas(self, event=None):
        if self.mca2 is None:
            return
        self.mca, self.mca2 = self.mca2, self.mca
        fore, back = self.det_fore, self.det_back
        self.wids['bkg_det'].SetSelection(fore)
        self.onSelectDet(index=back)

    def onMcaSumChoice(self, event=None):
        wid = self.wids['mca_sum']
        self.det.set_usesum('accum' in wid.GetStringSelection().lower())

    def onSetDwelltime(self, event=None, **kws):
        if 'dwelltime' in self.wids:
            self.det.set_dwelltime(dtime=self.wids['dwelltime'].GetValue())

    def clear_mcas(self):
        self.mca = self.mca2 = None
        self.x2data = self.y2data = None
        self.needs_newplot = True

    def onStart(self, event=None, dtime=None, **kws):
        if dtime is not None:
            self.wids['dwelltime'].SetValue("%.1f" % dtime)
            self.det.set_dwelltime(dtime=dtime)
        else:
            self.det.set_dwelltime(dtime=self.wids['dwelltime'].GetValue())
        self.det.start()

    def onStop(self, event=None, **kws):
        self.det.stop()
        self.det.needs_refresh = True
        time.sleep(0.05)
        self.UpdateData(event=None, force=True)

    def onErase(self, event=None, **kws):
        self.needs_newplot = True
        self.det.erase()

    def onDelROI(self, event=None):
        roiname = self.get_roiname()
        errmsg = None
        t0 = time.time()
        if self.roilist_sel is None:
            errmsg = 'No ROI selected to delete.'
        if errmsg is not None:
            return Popup(self, errmsg, 'Cannot Delete ROI')

        self.det.del_roi(roiname)
        XRFDisplayFrame.onDelROI(self)


    def onNewROI(self, event=None):
        roiname = self.get_roiname()
        errmsg = None
        if self.xmarker_left is None or self.xmarker_right is None:
            errmsg = 'Must select right and left markers to define ROI'
        elif roiname in self.wids['roilist'].GetStrings():
            errmsg = '%s is already in ROI list - use a unique name.' % roiname
        if errmsg is not None:
            return Popup(self, errmsg, 'Cannot Define ROI')

        confirmed = XRFDisplayFrame.onNewROI(self)
        if confirmed:
            self.det.add_roi(roiname, lo=self.xmarker_left,
                             hi=self.xmarker_right)

    def onRenameROI(self, event=None):
        roiname = self.get_roiname()
        errmsg = None
        if roiname in self.wids['roilist'].GetStrings():
            errmsg = '%s is already in ROI list - use a unique name.' % roiname
        elif self.roilist_sel is None:
            errmsg = 'No ROI selected to rename.'
        if errmsg is not None:
            return Popup(self, errmsg, 'Cannot Rename ROI')

        if self.roilist_sel < len(self.det.mcas[0].rois):
            self.det.rename_roi(self.roilist_sel, roiname)
            names = self.wids['roilist'].GetStrings()
            names[self.roilist_sel] = roiname
            self.wids['roilist'].Clear()
            for sname in names:
                self.wids['roilist'].Append(sname)
            self.wids['roilist'].SetSelection(self.roilist_sel)

    def onCalibrateEnergy(self, event=None, **kws):
        try:
            self.win_calib.Raise()
        except:
            self.win_calib = XRFCalibrationFrame(self, mca=self.mca,
                                              larch=self.larch,
                                              callback=self.onSetCalib)

    def onSetCalib(self, offset, slope, mca=None):
        print('XRFControl Set Energy Calibratione' , offset, slope, mca)

    def onClose(self, event=None):
        self.onStop()
        XRFDisplayFrame.onClose(self)

    def onExit(self, event=None):
        self.onStop()
        XRFDisplayFrame.onExit(self)

class EpicsXRFApp(wx.App, wx.lib.mixins.inspection.InspectionMixin):
    def __init__(self, **kws):
        self.kws = kws
        wx.App.__init__(self)

    def OnInit(self):
        self.Init()
        frame = EpicsXRFDisplayFrame(**self.kws) #
        frame.Show()
        self.SetTopWindow(frame)
        return True

if __name__ == "__main__":
    # e = EpicsXRFApp(prefix='QX4:', det_type='ME-4',
    #                amp_type='xspress3', nmca=4)

    EpicsXRFApp().MainLoop()



#!/usr/bin/env python
'''
Diffraction functions require for fitting and analyzing data.

mkak 2017.02.06 (originally written spring 2016)
'''

##########################################################################
# IMPORT PYTHON PACKAGES

import numpy as np

##########################################################################
# FUNCTIONS

def xy_file_reader(xyfile):
    '''
    Parses (x,y) data from xy text file.

    options:
    char - chararacter separating columns in data file (e.g. ',')
    '''

    units = None
    x, y = [], []
    with open(xyfile) as f:
        for line in f.readlines():
            if '#' not in line:
                fields = line.split()
                x += [float(fields[0])]
                y += [float(fields[1])]
            else:
                for opt in ['2th_deg','q_A^-1']:
                    if opt in line: units = opt

    return np.array(x),np.array(y),units

##########################################################################
def xy_file_writer(a,b,filename,char=None):

    if char:
        str = '%s' + char + '%s\n'
    else:
        str = '%s %s\n'

    with open(filename,'w') as f:
        for i,ai in a:
            f.write(str % (ai,b[i]))

    return()



#!/usr/bin/python
"""
subclass of wxmplot.ImageFrame specific for Map Viewer -- adds custom menus
"""

import os
import time
from threading import Thread
import socket

from functools import partial
import wx
try:
    from wx._core import PyDeadObjectError
except:
    PyDeadObjectError = Exception

is_wxPhoenix = 'phoenix' in wx.PlatformInfo

import numpy as np
from matplotlib.figure import Figure
from matplotlib.backends.backend_wxagg import FigureCanvasWxAgg

import wxmplot
from wxmplot import ImageFrame, PlotFrame, PlotPanel
from wxmplot.imagepanel import ImagePanel

from wxmplot.imageconf import ColorMap_List, Interp_List
from wxmplot.colors import rgb2hex, register_custom_colormaps

from wxutils import (SimpleText, TextCtrl, Button, Popup, Choice, pack)

def get_wxmplot_version():
    return "this function is deprecated"

class MapImageFrame(ImageFrame):
    """
    MatPlotlib Image Display on a wx.Frame, using ImagePanel
    """

    def __init__(self, parent=None, size=(650, 650), mode='intensity',
                 lasso_callback=None, move_callback=None, save_callback=None,
                 show_xsections=False, cursor_labels=None,
                 with_savepos=True,output_title='Image', **kws):
        self.det = None
        self.xrmfile = None
        self.map = None
        self.move_callback = move_callback
        self.save_callback = save_callback
        self.with_savepos = with_savepos

        ImageFrame.__init__(self, parent=parent, size=size,
                            lasso_callback=lasso_callback,
                            cursor_labels=cursor_labels, mode=mode,
                            output_title=output_title, **kws)

        w0, h0 = self.GetSize()
        w1, h1 = self.GetBestSize()
        self.SetSize((max(w0, w1)+5, max(h0, h1)+5))
        self.SetMinSize((500, 500))

        self.prof_plotter = None
        self.zoom_ini =  None
        self.lastpoint = [None, None]

        self.rbbox = None

    def display(self, map, det=None, xrmfile=None, xoff=0, yoff=0,
                with_savepos=True, **kws):
        self.xoff = xoff
        self.yoff = yoff
        self.det = det
        self.xrmfile = xrmfile
        self.map = map
        self.title = ''
        if 'title' in kws:
            self.title = kws['title']
        if 'contrast_level' not in kws:
            kws['contrast_level'] = 0.5
        ImageFrame.display(self, map, **kws)
        # self.set_contrast_levels()

        if self.save_callback is not None and hasattr(self, 'pos_name'):
            self.pos_name.Enable(with_savepos)

        sd = kws.get('subtitles', None)
        if sd is not None:
            for i, name in enumerate(('red', 'green', 'blue')):
                sub = sd.get(name, None)
                if sub is not None:
                    self.cmap_panels[i].title.SetLabel(sub)


    def onCursorMode(self, event=None, mode='zoom'):
        self.panel.cursor_mode = mode
        choice = self.zoom_mode.GetString(self.zoom_mode.GetSelection())
        if event is not None:
            if choice.startswith('Pick Area'):
                self.panel.cursor_mode = 'lasso'

    def onLasso(self, data=None, selected=None, mask=None, **kws):
        if hasattr(self.lasso_callback , '__call__'):

            self.lasso_callback(data=data, selected=selected, mask=mask,
                                xoff=self.xoff, yoff=self.yoff, det=self.det,
                                xrmfile=self.xrmfile, **kws)

        self.zoom_mode.SetSelection(0)
        self.panel.cursor_mode = 'zoom'

    def CustomConfig(self, panel, sizer=None, irow=0):
        """config panel for left-hand-side of frame"""

        labstyle = wx.ALIGN_LEFT|wx.LEFT|wx.TOP|wx.EXPAND

        if self.lasso_callback is None:
            zoom_opts = ('Zoom to Rectangle',)
        else:
            zoom_opts = ('Zoom to Rectangle',
                         'Pick Area for XRF Spectrum')

        cpanel = wx.Panel(panel)
        if sizer is None:
            sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(SimpleText(cpanel, label='Cursor Modes', style=labstyle),
                  0, labstyle, 3)
        self.zoom_mode = wx.RadioBox(cpanel, -1, "",
                                     wx.DefaultPosition, wx.DefaultSize,
                                     zoom_opts, 1, wx.RA_SPECIFY_COLS)
        self.zoom_mode.Bind(wx.EVT_RADIOBOX, self.onCursorMode)

        sizer.Add(self.zoom_mode, 1, labstyle, 4)

        if self.save_callback is not None:
            sizer.Add(SimpleText(cpanel, label='Save Position:', style=labstyle),
                      0, labstyle, 3)
            self.pos_name = wx.TextCtrl(cpanel, -1, '',  size=(155, -1),
                                        style=wx.TE_PROCESS_ENTER)
            self.pos_name.Bind(wx.EVT_TEXT_ENTER, self.onSavePixel)
            sizer.Add(self.pos_name, 0, labstyle, 3)

        pack(cpanel, sizer)
        return cpanel

    def onSavePixel(self, event=None):
        ix, iy = self.panel.conf.slice_xy
        if ix > 0 and iy > 0 and self.save_callback is not None:
            name  = str(event.GetString().strip())
            x = float(self.panel.xdata[int(ix)])
            y = float(self.panel.ydata[int(iy)])
            self.save_callback(name, ix, iy, x=x, y=y,
                               title=self.title, xrmfile=self.xrmfile)


class CorrelatedMapFrame(wxmplot.ImageMatrixFrame):
    """
    wx.Frame, with 3 ImagePanels and correlation plot for 2 map arrays
    """
    def __init__(self, parent=None, xrmfile=None,
                 title='CorrelatedMaps',  **kws):

        self.xrmfile = xrmfile
        wxmplot.ImageMatrixFrame.__init__(self, parent, size=(900, 625),
                                          title=title, **kws)

    def CustomConfig(self, parent):
        pass


class with_profile_mode:
    def prof_motion(self, event=None):
        if not event.inaxes or self.zoom_ini is None:
            return
        try:
            xmax, ymax  = event.x, event.y
        except:
            return
        xmin, ymin, xd, yd = self.zoom_ini
        if event.xdata is not None:
            self.lastpoint[0] = event.xdata
        if event.ydata is not None:
            self.lastpoint[1] = event.ydata

        yoff = self.panel.canvas.figure.bbox.height
        ymin, ymax = yoff - ymin, yoff - ymax

        zdc = wx.ClientDC(self.panel.canvas)
        zdc.SetLogicalFunction(wx.XOR)
        zdc.SetBrush(wx.TRANSPARENT_BRUSH)
        zdc.SetPen(wx.Pen('White', 2, wx.SOLID))
        zdc.ResetBoundingBox()
        if not is_wxPhoenix:
            zdc.BeginDrawing()

        # erase previous box
        if self.rbbox is not None:
            zdc.DrawLine(*self.rbbox)
        self.rbbox = (xmin, ymin, xmax, ymax)
        zdc.DrawLine(*self.rbbox)
        if not is_wxPhoenix:
            zdc.EndDrawing()

    def prof_leftdown(self, event=None):
        self.report_leftdown(event=event)
        if event.inaxes: #  and len(self.map.shape) == 2:
            self.lastpoint = [None, None]
            self.zoom_ini = [event.x, event.y, event.xdata, event.ydata]

    def prof_leftup(self, event=None):
        # print("Profile Left up ", self.map.shape, self.rbbox)
        if self.rbbox is not None:
            zdc = wx.ClientDC(self.panel.canvas)
            zdc.SetLogicalFunction(wx.XOR)
            zdc.SetBrush(wx.TRANSPARENT_BRUSH)
            zdc.SetPen(wx.Pen('White', 2, wx.SOLID))
            zdc.ResetBoundingBox()
            if not is_wxPhoenix:
                zdc.BeginDrawing()
            zdc.DrawLine(*self.rbbox)
            if not is_wxPhoenix:
                zdc.EndDrawing()
            self.rbbox = None

        if self.zoom_ini is None or self.lastpoint[0] is None:
            return

        x0 = int(self.zoom_ini[2])
        x1 = int(self.lastpoint[0])
        y0 = int(self.zoom_ini[3])
        y1 = int(self.lastpoint[1])
        dx, dy = abs(x1-x0), abs(y1-y0)

        self.lastpoint, self.zoom_ini = [None, None], None
        if dx < 2 and dy < 2:
            self.zoom_ini = None
            return

        outdat = []
        if dy > dx:
            _y0 = min(int(y0), int(y1+0.5))
            _y1 = max(int(y0), int(y1+0.5))

            for iy in range(_y0, _y1):
                ix = int(x0 + (iy-int(y0))*(x1-x0)/(y1-y0))
                outdat.append((ix, iy))
        else:
            _x0 = min(int(x0), int(x1+0.5))
            _x1 = max(int(x0), int(x1+0.5))
            for ix in range(_x0, _x1):
                iy = int(y0 + (ix-int(x0))*(y1-y0)/(x1-x0))
                outdat.append((ix, iy))
        x, y, z = [], [], []
        for ix, iy in outdat:
            x.append(ix)
            y.append(iy)
            z.append(self.panel.conf.data[iy, ix])
        self.prof_dat = dy>dx, outdat

        if self.prof_plotter is not None:
            try:
                self.prof_plotter.Raise()
                self.prof_plotter.clear()

            except (AttributeError, PyDeadObjectError):
                self.prof_plotter = None

        if self.prof_plotter is None:
            self.prof_plotter = PlotFrame(self, title='Profile')
            self.prof_plotter.panel.report_leftdown = self.prof_report_coords

        xlabel, y2label = 'Pixel (x)',  'Pixel (y)'

        x = np.array(x)
        y = np.array(y)
        z = np.array(z)
        if dy > dx:
            x, y = y, x
            xlabel, y2label = y2label, xlabel
        self.prof_plotter.panel.clear()

        if len(self.title) < 1:
            self.title = os.path.split(self.xrmfile.filename)[1]

        opts = dict(linewidth=2, marker='+', markersize=3,
                    show_legend=True, xlabel=xlabel)

        if isinstance(z[0], np.ndarray) and len(z[0]) == 3: # color plot
            rlab = self.subtitles['red']
            glab = self.subtitles['green']
            blab = self.subtitles['blue']
            self.prof_plotter.plot(x, z[:, 0], title=self.title, color='red',
                                   zorder=20, xmin=min(x)-3, xmax=max(x)+3,
                                   ylabel='counts', label=rlab, **opts)
            self.prof_plotter.oplot(x, z[:, 1], title=self.title, color='darkgreen',
                                   zorder=20, xmin=min(x)-3, xmax=max(x)+3,
                                   ylabel='counts', label=glab, **opts)
            self.prof_plotter.oplot(x, z[:, 2], title=self.title, color='blue',
                                   zorder=20, xmin=min(x)-3, xmax=max(x)+3,
                                   ylabel='counts', label=blab, **opts)

        else:

            self.prof_plotter.plot(x, z, title=self.title, color='blue',
                                   zorder=20, xmin=min(x)-3, xmax=max(x)+3,
                                   ylabel='counts', label='counts', **opts)

        self.prof_plotter.oplot(x, y, y2label=y2label, label=y2label,
                              zorder=3, side='right', color='black', **opts)

        self.prof_plotter.panel.unzoom_all()
        self.prof_plotter.Show()
        self.zoom_ini = None

        self.zoom_mode.SetSelection(0)
        self.panel.cursor_mode = 'zoom'

    def prof_report_coords(self, event=None):
        """override report leftdown for profile plotter"""
        if event is None:
            return
        ex, ey = event.x, event.y
        msg = ''
        plotpanel = self.prof_plotter.panel
        axes  = plotpanel.fig.properties()['axes'][0]
        write = plotpanel.write_message
        try:
            x, y = axes.transData.inverted().transform((ex, ey))
        except:
            x, y = event.xdata, event.ydata

        if x is None or y is None:
            return

        _point = 0, 0, 0, 0, 0
        for ix, iy in self.prof_dat[1]:
            if (int(x) == ix and not self.prof_dat[0] or
                int(x) == iy and self.prof_dat[0]):
                _point = (ix, iy,
                              self.panel.xdata[ix],
                              self.panel.ydata[iy],
                              self.panel.conf.data[iy, ix])

        msg = "Pixel [%i, %i], X, OME = [%.4f mm, %.4f deg], Intensity= %g" % _point
        write(msg,  panel=0)



#!/usr/bin/env python
"""
Helper classes for larch interpreter
"""
from __future__ import division
import sys, os, time
import ast
import numpy as np
import traceback
import inspect
from collections import OrderedDict
import ctypes
import ctypes.util

from .symboltable import Group, isgroup
from .site_config import usr_larchdir
from .closure import Closure
from .utils import uname, bindir

HAS_TERMCOLOR = False
try:
    from termcolor import colored
    if uname == 'win':
        # HACK (hopefully temporary):
        # disable color output for Windows command terminal
        # because it interferes with wx event loop.
        import CannotUseTermcolorOnWindowsWithWx
        # os.environ.pop('TERM')
        # import colorama
        # colorama.init()
    HAS_TERMCOLOR = True
except ImportError:
    HAS_TERMCOLOR = False

HAS_YAML = False
try:
    import yaml
    HAS_YAML = True
except ImportError:
    HAS_YAML = False

class LarchPluginException(Exception):
    """Exception with Larch Plugin"""
    def __init__(self, msg):
        Exception.__init__(self)
        self.msg = msg

    def __str__(self):
        return "\n%s" % (self.msg)

class Empty:
    def __nonzero__(self): return False

# holder for 'returned None' from Larch procedure
ReturnedNone = Empty()

def get_filetext(fname, lineno):
    """try to extract line from source text file"""
    out = '<could not find text>'
    try:
        ftmp = open(fname, 'r')
        lines = ftmp.readlines()
        ftmp.close()
        lineno = min(lineno, len(lines)) - 1
        out = lines[lineno][:-1]
    except:
        pass
    return out

class LarchExceptionHolder:
    "basic exception handler"
    def __init__(self, node=None, msg='', fname='<stdin>',
                 func=None, expr=None, exc=None, lineno=0):
        self.node = node
        self.fname  = fname
        self.func = func
        self.expr = expr
        self.msg  = msg
        self.exc  = exc
        self.lineno = lineno
        self.exc_info = sys.exc_info()

        if self.exc is None and self.exc_info[0] is not None:
            self.exc = self.exc_info[0]
        if self.msg in ('', None) and self.exc_info[1] is not None:
            self.msg = self.exc_info[1]

    def get_error(self):
        "retrieve error data"
        col_offset = -1
        e_type, e_val, e_tb = self.exc_info
        if self.node is not None:
            try:
                col_offset = self.node.col_offset
            except AttributeError:
                pass
        try:
            exc_name = self.exc.__name__
        except AttributeError:
            exc_name = str(self.exc)
        if exc_name in (None, 'None'):
            exc_name = 'UnknownError'

        out = []
        fname = self.fname

        if isinstance(self.expr, ast.AST):
            self.expr = 'In compiled script'
        if self.expr is None:
            out.append('unknown error\n')
        elif '\n' in self.expr:
            out.append("\n%s" % self.expr)
        else:
            out.append("    %s" % self.expr)
        if col_offset > 0:
            out.append("%s^^^" % ((col_offset)*' '))

        fline = '   File %s, line %i' % (fname, self.lineno)
        if self.func is not None:
            func = self.func
            fname = self.fname
            if fname is None:
                if isinstance(func, Closure):
                    func = func.func
                    fname = inspect.getmodule(func).__file__
                try:
                    fname = inspect.getmodule(func).__file__
                except AttributeError:
                    fname = 'unknown'
            if fname.endswith('.pyc'):
                fname = fname[:-1]

            if hasattr(self.func, 'name'):
                dec = ''
                if isinstance(self.func, Procedure):
                    dec = 'procedure '
                pname = self.func.name
                ftext = get_filetext(self.fname, self.lineno)
                fline = "%s, in %s%s\n%s" % (fline, dec, pname, ftext)

        if fline is not None:
            out.append(fline)

        tblist = []
        for tb in traceback.extract_tb(self.exc_info[2]):
            if not (sys.prefix in tb[0] and
                    ('ast.py' in tb[0] or
                     os.path.join('larch', 'utils') in tb[0] or
                     os.path.join('larch', 'interpreter') in tb[0] or
                     os.path.join('larch', 'symboltable') in tb[0])):
                tblist.append(tb)
        if len(tblist) > 0:
            out.append(''.join(traceback.format_list(tblist)))

        # try to get last error message, as from e_val.args
        ex_msg = getattr(e_val, 'args', None)
        try:
            ex_msg = ' '.join(ex_msg)
        except TypeError:
            pass

        if ex_msg is None:
            ex_msg = getattr(e_val, 'message', None)
        if ex_msg is None:
            ex_msg = self.msg
        out.append("%s: %s" % (exc_name, ex_msg))

        out.append("")
        return (exc_name, '\n'.join(out))



class StdWriter(object):
    """Standard writer method for Larch,
    to be used in place of sys.stdout

    supports methods:
      set_mode(mode) # one of 'text', 'text2', 'error', 'comment'
      write(text)
      flush()
    """
    valid_termcolors = ('grey', 'red', 'green', 'yellow',
                        'blue', 'magenta', 'cyan', 'white')

    termcolor_attrs = ('bold', 'underline', 'blink', 'reverse')
    def __init__(self, stdout=None, has_color=True, _larch=None):
        if stdout is None:
            stdout = sys.stdout
        self.has_color = has_color and HAS_TERMCOLOR
        self.writer = stdout
        self._larch = _larch
        self.textstyle = None

    def set_textstyle(self, mode='text'):
        """ set text style for output """
        if not self.has_color:
            self.textstyle = None
        display_colors = self._larch.symtable._sys.display.colors
        self.textstyle =  display_colors.get(mode, {})

    def write(self, text):
        """write text to writer
        write('hello')
        """
        if self.textstyle is not None and HAS_TERMCOLOR:
            text = colored(text, **self.textstyle)
        self.writer.write(text)

    def flush(self):
        self.writer.flush()


class Procedure(object):
    """larch procedure:  function """
    def __init__(self, name, _larch=None, doc=None,
                 fname='<stdin>', lineno=0,
                 body=None, args=None, kwargs=None,
                 vararg=None, varkws=None):
        self.name     = name
        self._larch    = _larch
        self.modgroup = _larch.symtable._sys.moduleGroup
        self.body     = body
        self.argnames = args
        self.kwargs   = kwargs
        self.vararg   = vararg
        self.varkws   = varkws
        self.__doc__  = doc
        self.lineno   = lineno
        self.__file__ = fname
        self.__name__ = name

    def __repr__(self):
        return "<Procedure %s, file=%s>" % (self.name, self.__file__)

    def _signature(self):
        sig = ""
        if len(self.argnames) > 0:
            sig = "%s%s" % (sig, ', '.join(self.argnames))
        if self.vararg is not None:
            sig = "%s, *%s" % (sig, self.vararg)
        if len(self.kwargs) > 0:
            if len(sig) > 0:
                sig = "%s, " % sig
            _kw = ["%s=%s" % (k, repr(v)) for k, v in self.kwargs]
            sig = "%s%s" % (sig, ', '.join(_kw))

            if self.varkws is not None:
                sig = "%s, **%s" % (sig, self.varkws)
        return "%s(%s)" % (self.name, sig)

    def raise_exc(self, **kws):
        ekws = dict(lineno=self.lineno, func=self, fname=self.__file__)
        ekws.update(kws)
        self._larch.raise_exception(None,  **ekws)

    def __call__(self, *args, **kwargs):
        # msg = 'Cannot run Procedure %s' % self.name
        lgroup  = Group()
        lgroup.__name__ = hex(id(lgroup))
        args    = list(args)
        nargs  = len(args)
        nkws   = len(kwargs)
        nargs_expected = len(self.argnames)


        # case 1: too few arguments, but the correct keyword given
        if (nargs < nargs_expected) and nkws > 0:
            for name in self.argnames[nargs:]:
                if name in kwargs:
                    args.append(kwargs.pop(name))
            nargs = len(args)
            nargs_expected = len(self.argnames)
            nkws = len(kwargs)

        # case 2: multiple values for named argument
        if len(self.argnames) > 0 and kwargs is not None:
            msg = "%s() got multiple values for keyword argument '%s'"
            for targ in self.argnames:
                if targ in kwargs:
                    self.raise_exc(exc=TypeError,
                                   msg=msg % (self.name, targ))
                    return

        # case 3: too few args given
        if nargs < nargs_expected:
            mod = 'at least'
            if len(self.kwargs) == 0:
                mod = 'exactly'
            msg = '%s() expected %s %i arguments (got %i)'
            self.raise_exc(exc=TypeError,
                           msg=msg%(self.name, mod, nargs_expected, nargs))
            return

        # case 4: more args given than expected, varargs not given
        if nargs > nargs_expected and self.vararg is None:
            if nargs - nargs_expected > len(self.kwargs):
                msg = 'too many arguments for %s() expected at most %i, got %i'
                msg = msg % (self.name, len(self.kwargs)+nargs_expected, nargs)
                self.raise_exc(exc=TypeError, msg=msg)
                return
            for i, xarg in enumerate(args[nargs_expected:]):
                kw_name = self.kwargs[i][0]
                if kw_name not in kwargs:
                    kwargs[kw_name] = xarg

        for argname in self.argnames:
            if len(args) > 0:
                setattr(lgroup, argname, args.pop(0))
        try:
            if self.vararg is not None:
                setattr(lgroup, self.vararg, tuple(args))

            for key, val in self.kwargs:
                if key in kwargs:
                    val = kwargs.pop(key)
                setattr(lgroup, key, val)

            if self.varkws is not None:
                setattr(lgroup, self.varkws, kwargs)
            elif len(kwargs) > 0:
                msg = 'extra keyword arguments for procedure %s (%s)'
                msg = msg % (self.name, ','.join(list(kwargs.keys())))
                self.raise_exc(exc=TypeError, msg=msg)
                return

        except (ValueError, LookupError, TypeError,
                NameError, AttributeError):
            msg = 'incorrect arguments for procedure %s' % self.name
            self.raise_exc(msg=msg)
            return

        stable  = self._larch.symtable
        stable.save_frame()
        stable.set_frame((lgroup, self.modgroup))
        retval = None
        self._larch.retval = None
        self._larch.debug = True
        for node in self.body:
            self._larch.run(node, fname=self.__file__, func=self,
                            lineno=node.lineno+self.lineno-1, with_raise=False)
            if len(self._larch.error) > 0:
                break
            if self._larch.retval is not None:
                retval = self._larch.retval
                if retval is ReturnedNone: retval = None
                break
        stable.restore_frame()
        self._larch.debug = False
        self._larch.retval = None
        del lgroup
        return retval

def enable_plugins():
    """add all available Larch plugin paths
    """
    if 'larch_plugins' not in sys.modules:
        import larch
        sys.modules['larch_plugins'] = larch
    return sys.modules['larch_plugins']

def add2path(envvar='PATH', dirname='.'):
    """add specified dir to begninng of PATH and
    DYLD_LIBRARY_PATH, LD_LIBRARY_PATH environmental variables,
    returns previous definition of PATH, for restoration"""
    sep = ':'
    if uname == 'win':
        sep = ';'
    oldpath = os.environ.get(envvar, '')
    if oldpath == '':
        os.environ[envvar] = dirname
    else:
        paths = oldpath.split(sep)
        paths.insert(0, os.path.abspath(dirname))
        os.environ[envvar] = sep.join(paths)
    return oldpath


def isNamedClass(obj, cls):
    """this is essentially a replacement for
      isinstance(obj, cls)
    that looks if an objects class name matches that of a class
    obj.__class__.__name__ == cls.__name__
    """
    return  obj.__class__.__name__ == cls.__name__

def get_dll(libname):
    """find and load a shared library"""
    _dylib_formats = {'win': '%s.dll', 'linux': 'lib%s.so',
                      'darwin': 'lib%s.dylib'}

    loaddll = ctypes.cdll.LoadLibrary
    if uname == 'win':
        loaddll = ctypes.windll.LoadLibrary

    # normally, we expect the dll to be here in the larch dlls tree
    # if we find it there, use that one
    fname = _dylib_formats[uname] % libname
    dllpath = os.path.join(bindir, fname)
    if os.path.exists(dllpath):
        return loaddll(dllpath)

    # if not found in the larch dlls tree, try your best!
    dllpath = ctypes.util.find_library(libname)
    if dllpath is not None and os.path.exists(dllpath):
        return loaddll(dllpath)
    return None


def read_workdir(conffile):
    """read working dir from a config file in the users larch dir
    compare save_workdir(conffile) which will save this value

    can be used to ensure that application startup starts in
    last working directory
    """

    try:
        w_file = os.path.join(usr_larchdir, conffile)
        if os.path.exists(w_file):
            line = open(w_file, 'r').readlines()
            workdir = line[0][:-1]
            os.chdir(workdir)
    except:
        pass

def save_workdir(conffile):
    """write working dir to a config file in the users larch dir
    compare read_workdir(conffile) which will read this value

    can be used to ensure that application startup starts in
    last working directory
    """

    try:
        w_file = os.path.join(usr_larchdir, conffile)
        fh = open(w_file, 'w')
        fh.write("%s\n" % os.getcwd())
        fh.close()
    except:
        pass


def read_config(conffile):
    """read yaml config file from users larch dir
    compare save_config(conffile) which will save such a config

    returns dictionary / configuration
    """
    cfile = os.path.join(usr_larchdir, conffile)
    out = None
    if os.path.exists(cfile):
        with open(cfile, 'r') as fh:
            out = fh.read()
    if out is not None:
        if not HAS_YAML:
            raise RuntimeError('yaml is not available')
        out = yaml.load(out)
    return out

def save_config(conffile, config):
    """write yaml config file in the users larch dir
    compare read_confif(conffile) which will read this value

    """
    cfile = os.path.join(usr_larchdir, conffile)
    if not HAS_YAML:
        raise RuntimeError('yaml is not available')
    out = yaml.dump(config)
    with open(cfile, 'w') as fh:
        fh.write(out)

def parse_group_args(arg0, members=None, group=None, defaults=None,
                     fcn_name=None, check_outputs=True):
    """parse arguments for functions supporting First Argument Group convention

    That is, if the first argument is a Larch Group and contains members
    named in 'members', this will return data extracted from that group.

    Arguments
    ----------
    arg0:         first argument for function call.
    members:      list/tuple of names of required members (in order)
    defaults:     tuple of default values for remaining required
                  arguments past the first (in order)
    group:        group sent to parent function, used for outputs
    fcn_name:     name of parent function, used for error messages
    check_output: True/False (default True) setting whether a Warning should
                  be raised in any of the outputs (except for the final group)
                  are None.  This effectively checks that all expected inputs
                  have been specified
    Returns
    -------
     tuple of output values in the order listed by members, followed by the
     output group (which could be None).

    Notes
    -----
    This implements the First Argument Group convention, used for many Larch functions.
    As an example, the function _xafs.find_e0 is defined like this:
       find_e0(energy, mu=None, group=None, ...)

    and uses this function as
       energy, mu, group = parse_group_arg(energy, members=('energy', 'mu'),
                                           defaults=(mu,), group=group,
                                           fcn_name='find_e0', check_output=True)

    This allows the caller to use
         find_e0(grp)
    as a shorthand for
         find_e0(grp.energy, grp.mu, group=grp)

    as long as the Group grp has member 'energy', and 'mu'.

    With 'check_output=True', the value for 'mu' is not actually allowed to be None.

    The defaults tuple should be passed so that correct values are assigned
    if the caller actually specifies arrays as for the full call signature.
    """
    if members is None:
        members = []
    if isgroup(arg0, *members):
        if group is None:
            group = arg0
        out = [getattr(arg0, attr) for attr in members]
    else:
        out = [arg0] + list(defaults)

    # test that all outputs are non-None
    if check_outputs:
        _errmsg = """%s: needs First Argument Group or valid arguments for
  %s"""
        if fcn_name is None:
            fcn_name ='unknown function'
        for i, nam in enumerate(members):
            if out[i] is None:
                raise Warning(_errmsg % (fcn_name, ', '.join(members)))

    out.append(group)
    return out

def Make_CallArgs(skipped_args):
    """
    decorator to create a 'call_args' dictionary
    containing function arguments
    """
    def wrap(fcn):
        def wrapper(*args, **kwargs):
            result = fcn(*args, **kwargs)
            argspec = inspect.getfullargspec(fcn)

            offset = len(argspec.args) - len(argspec.defaults)
            call_args = OrderedDict()

            for k in argspec.args[:offset]:
                call_args[k] = None
            for k, v in zip(argspec.args[offset:], argspec.defaults):
                call_args[k] = v

            for iarg, arg in enumerate(args):
                call_args[argspec.args[iarg]] = arg

            call_args.update(kwargs)

            skipped = skipped_args[:]
            at0 = skipped[0]
            at1 = skipped[1]
            a, b, groupx = parse_group_args(call_args[at0],
                                            members=(at0, at1),
                                            defaults=(call_args[at1],),
                                            group=call_args['group'],
                                            fcn_name=fcn.__name__)

            for k in skipped + ['group', '_larch']:
                if k in call_args:
                    call_args.pop(k)

            details_name = '%s_details' % fcn.__name__
            if not hasattr(groupx, details_name):
                setattr(groupx, details_name, Group())
            setattr(getattr(groupx, details_name),
                    'call_args', call_args)
            return result
        wrapper.__doc__ = fcn.__doc__
        wrapper.__name__ = fcn.__name__
        wrapper._larchfunc_ = fcn
        wrapper.__filename__ = fcn.__code__.co_filename
        wrapper.__dict__.update(fcn.__dict__)
        return wrapper
    return wrap

def ValidateLarchPlugin(fcn):
    """function decorator to ensure that _larch is included in keywords,
    and that it is a valid Interpeter in that it has:
    1. a symtable attribute
    2. a writer attribute
    """
    errmsg1 = "plugin function '%s' needs a '_larch' argument"
    errmsg2 = "plugin function '%s' has an invalid '_larch'  '%s'"

    def wrapper(*args, **keywords):
        "ValidateLarchPlugin"
        _larch = keywords.get('_larch', None)
        if _larch is None:
            raise LarchPluginException(errmsg1 % fcn.__name__)

        symtab = getattr(_larch, 'symtable', None)
        writer = getattr(_larch, 'writer', None)
        if not (isgroup(symtab) and callable(getattr(writer, 'write', None))):
            raise LarchPluginException(errmsg2 % (fcn.__name__, _larch))
        return fcn(*args, **keywords)
    wrapper.__doc__ = fcn.__doc__
    wrapper.__name__ = fcn.__name__
    wrapper._larchfunc_ = fcn
    wrapper.__filename__ = fcn.__code__.co_filename
    wrapper.__dict__.update(fcn.__dict__)
    return wrapper


def ensuremod(_larch, modname=None):
    "ensure that a group exists"
    if _larch is not None:
        symtable = _larch.symtable
        if modname is not None and not symtable.has_group(modname):
            symtable.newgroup(modname)
        return symtable



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Custom version of SILX Plot2D
=============================
"""
import time
import numpy as np
from silx.gui.plot import Plot2D as silxPlot2D
from larch.utils.logging import getLogger


class Plot2D(silxPlot2D):
    """Custom Plot2D instance targeted to 2D images"""

    def __init__(self, parent=None, backend=None, logger=None, title="Plot2D"):

        super(Plot2D, self).__init__(parent=parent, backend=backend)

        self._logger = logger or getLogger("Plot2D")
        self._index = None
        self._title = title
        self.setWindowTitle(self._title)
        self._image = None
        self._mask = None
        self._origin = (0, 0)
        self._scale = (1, 1)
        self._xlabel = 'X'
        self._ylabel = 'Y'
        self.setKeepDataAspectRatio(True)
        self.getDefaultColormap().setName('viridis')

    def _drawContours(self, values, color='gray',
                      plot_timeout=100, plot_method='curve'):
        """Draw iso contours for given values

        Parameters
        ----------
        values : list or array
            intensities at which to find contours
        color : string (optional)
            color of contours (among common color names) ['gray']
        plot_timeout : int (optional)
            time in seconds befor the plot is interrupted
        plot_method : str (optional)
            method to use for the contour plot
            'curve' -> self.addCurve, polygons as from find_contours
            'curve_max' -> self.addCurve, one polygon (max length)
            'curve_merge' -> self.addCurve, one polygon (concatenate)
            'scatter' -> self.addScatter (only points)
        """
        if self._ms is None:
            return
        ipolygon = 0
        totTime = 0
        for ivalue, value in enumerate(values):
            startTime = time.time()
            polygons = self._ms.find_contours(value)
            polTime = time.time()
            self._logger.debug(f"Found {len(polygons)} polygon at level {value}")
            totTime += polTime - startTime
            # prepare polygons list for plot_method
            if len(polygons) == 0:
                continue
            if len(polygons) > 1:
                if (plot_method == 'curve_max'):
                    from sloth.utils.arrays import imax
                    lengths = [len(x) for x in polygons]
                    polygons = [polygons[imax(lengths)]]
                elif (plot_method == 'curve_merge') or (plot_method == 'scatter'):
                    polygons = [np.concatenate(polygons, axis=0)]
                else:
                    pass
            # define default contour style
            contourStyle = {"linestyle": "-",
                            "linewidth": 0.5,
                            "color": color}
            for polygon in polygons:
                legend = "polygon-%d" % ipolygon
                xpoly = polygon[:, 1]
                ypoly = polygon[:, 0]
                xscale = np.ones_like(xpoly) * self._scale[0]
                yscale = np.ones_like(ypoly) * self._scale[1]
                xorigin = np.ones_like(xpoly) * self._origin[0]
                yorigin = np.ones_like(ypoly) * self._origin[1]
                x = xpoly * xscale + xorigin
                y = ypoly * yscale + yorigin
                # plot timeout
                if totTime >= plot_timeout:
                    self._logger.warning("Plot contours time out reached!")
                    break
                # plot methods
                if plot_method == 'scatter':
                    from silx.gui.colors import (Colormap, rgba)
                    cm = Colormap()
                    cm.setColormapLUT([rgba(color)])
                    arrval = np.ones_like(x)*value
                    self.addScatter(x, y, arrval, symbol='.', colormap=cm)
                else:
                    self.addCurve(x=x, y=y, legend=legend, resetzoom=False,
                                  **contourStyle)
                pltTime = time.time()
                totTime += pltTime - polTime
                ipolygon += 1

    def addContours(self, nlevels, algo='merge', **draw_kwars):
        """Add contour lines to plot

        Parameters
        ----------
        nlevels : int
            number of contour levels to plot

        algo : str (optional)
            marching squares algorithm implementation
            'merge' -> silx
            'skimage' -> scikit-image
        color : str, optional
            color of contour lines ['gray']
        linestyle : str, optional
            line style of contour lines ['-']
        linewidth : int, optional
            line width of contour lines [1]

        Returns
        -------
        None
        """
        image = self._image
        mask = self._mask
        if image is None:
            self._logger.error('add image first!')
        if algo == 'merge':
            from silx.image.marchingsquares._mergeimpl import MarchingSquaresMergeImpl
            self._ms = MarchingSquaresMergeImpl(image, mask=mask)
        elif algo == 'skimage':
            try:
                import skimage
                from silx.image.marchingsquares._skimage import MarchingSquaresSciKitImage
                self._ms = MarchingSquaresSciKitImage(image,
                                                      mask=mask)
            except ImportError:
                self._logger.error('skimage not found')
                self._ms = None
        else:
            self._ms = None
        imgmin, imgmax = image.min(), image.max()
        delta = (imgmax - imgmin) / nlevels
        values = np.arange(imgmin, imgmax, delta)
        self._drawContours(values, **draw_kwars)

    def index(self):
        if self._index is None:
            self._index = 0
        return self._index

    def setIndex(self, value):
        self._index = value
        if self._index is not None:
            self.setWindowTitle('{0}: {1}'.format(self._index, self._title))

    def reset(self):
        self.clear()
        self.setGraphTitle()
        self.setGraphXLabel('X')
        # self.setGraphXLimits(0, 100)
        self.setGraphYLabel('Y')
        # self.setGraphYLimits(0, 100)

    def addImage(self, data, x=None, y=None,
                 title=None, xlabel=None, ylabel=None,
                 vmin=None, vmax=None, **kwargs):
        """Custom addImage

        Parameters
        ----------
        data : array
        x, y : None or array (optional)
            x, y to set origin and scale (both should be given!)
        title : str
            set self.setGraphTitle(str) / self.setWindowTitle(str)
        xlabel, ylabel : None or str (optional)
            set self.setGraphXLabel / self.setGraphYLabel
        vmin, vmax : float (optional)
            intensity values of the colormap min/max
        """
        self._image = data
        self._x = x
        self._y = y
        if (x is not None) and (y is not None):
            self._origin = (np.min(x), np.min(y))
            self._scale = (x[1]-x[0], y[1]-y[0])
        if title is not None:
            self._title = title
            self.setGraphTitle(title)
            if self._index is not None:
                self.setWindowTitle('{0}: {1}'.format(self._index, self._title))
            else:
                self.setWindowTitle(self._title)
        if xlabel is not None:
            self._xlabel = xlabel
            self.setGraphXLabel(xlabel)
        if ylabel is not None:
            self._ylabel = ylabel
            self.setGraphYLabel(ylabel)
        if (vmin is None):
            vmin = self._image.min()
        if (vmax is None):
            vmax = self._image.max()
        self.getDefaultColormap().setVRange(vmin, vmax)
        return super(Plot2D, self).addImage(data, origin=self._origin,
                                            scale=self._scale,
                                            **kwargs)


def dummy_gauss_image(x=None, y=None,
                      xhalfrng=1.5, yhalfrng=None, xcen=0.5, ycen=0.9,
                      xnpts=1024, ynpts=None, xsigma=0.55, ysigma=0.25,
                      noise=0.3):
    """Create a dummy 2D Gaussian image with noise

    Parameters
    ----------
    x, y : 1D arrays (optional)
        arrays where to generate the image [None -> generated]
    xhalfrng : float (optional)
        half range of the X axis [1.5]
    yhalfrng : float or None (optional)
        half range of the Y axis [None -> xhalfrng]
    xcen : float (optional)
        X center [0.5]
    ycen : float (optional)
        Y center [0.9]
    xnpts : int (optional)
        number of points X [1024]
    ynpts : int or None (optional)
        number of points Y [None -> xnpts]
    xsigma : float (optional)
        sigma X [0.55]
    ysigma : float (optional)
        sigma Y [0.25]
    noise : float (optional)
        random noise level between 0 and 1 [0.3]

    Returns
    -------
    x, y : 1D arrays
    signal : 2D array
    """
    if yhalfrng is None:
        yhalfrng = xhalfrng
    if ycen is None:
        ycen = xcen
    if ynpts is None:
        ynpts = xnpts
    if x is None:
        x = np.linspace(xcen-xhalfrng, xcen+xhalfrng, xnpts)
    if y is None:
        y = np.linspace(ycen-yhalfrng, ycen+yhalfrng, ynpts)
    xx, yy = np.meshgrid(x, y)
    signal = np.exp(-((xx-xcen)**2 / (2*xsigma**2) +
                      ((yy-ycen)**2 / (2*ysigma**2))))
    # add noise
    signal += noise * np.random.random(size=signal.shape)
    return x, y, signal


def main(contour_levels=5, noise=0.1, compare_with_matplolib=False,
         plot_method='curve'):
    """Run a Qt app with the widget"""
    from silx import sx
    sx.enable_gui()
    xhalfrng = 10.5
    yhalfrng = 5.5
    npts = 1024
    xcen = 0
    ycen = 0
    x = np.linspace(xcen-0.7*xhalfrng, xcen+1.3*xhalfrng, npts)
    y = np.linspace(ycen-0.7*yhalfrng, ycen+1.3*yhalfrng, npts)
    x1, y1, signal1 = dummy_gauss_image(x=x, y=y, xcen=xcen, ycen=ycen,
                                        xsigma=3, ysigma=1.1,
                                        noise=noise)
    x2, y2, signal2 = dummy_gauss_image(x=x, y=y,
                                        xcen=4.2, ycen=2.2,
                                        xsigma=3, ysigma=2.1,
                                        noise=noise)
    signal = signal1 + 0.8*signal2
    p = Plot2D(backend='matplotlib')
    p.addImage(signal, x=x, y=y, xlabel='X', ylabel='Y')
    p.addContours(contour_levels, plot_method=plot_method)
    p.show()

    if compare_with_matplolib:
        import matplotlib.pyplot as plt
        from matplotlib import cm
        plt.ion()
        plt.close('all')
        fig, ax = plt.subplots()
        imgMin, imgMax = np.min(signal), np.max(signal)
        values = np.linspace(imgMin, imgMax, contour_levels)
        extent = (x.min(), x.max(), y.min(), y.max())
        ax.imshow(signal, origin='lower', extent=extent,
                  cmap=cm.viridis)
        ax.contour(x, y, signal, values, origin='lower', extent=extent,
                   colors='gray', linewidths=1)
        ax.set_title("pure matplotlib")
        plt.show()

    input("Press enter to close window")


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
  Larch column file reader: read_ascii
"""
import os
import sys
import time
import string
from collections import namedtuple
import numpy as np
from dateutil.parser import parse as dateparse
from math import log10
from larch import Group
from larch.symboltable import isgroup
from .fileutils import fix_varname

from .xafs_beamlines import guess_beamline

nanresult = namedtuple('NanResult', ('file_ok', 'message', 'nan_rows',
                                     'nan_cols', 'inf_rows', 'inf_cols'))


MODNAME = '_io'
TINY = 1.e-7
MAX_FILESIZE = 100*1024*1024  # 100 Mb limit
COMMENTCHARS = '#;%*!$'


def look_for_nans(path):
    """
    look for Nans and Infs in an ascii data file

    Returns
    -------
    NanResult with 'file_ok', 'message', 'nan_rows', 'nan_cols',
    'inf_rows', and 'inf_cols':

    'file_ok' : True if data is read and contains no Nans or Infs
    'file_ok' : False if path cannot be read or data contaings Nans / Infs
    'message' : exception message if file cannot be read at all or
                'has nans', 'has infs' or 'has nans and infs'
    nan_rows, etc:  list of rows, columns containing Nans or Infs
    """

    nan_rows, nan_cols, inf_rows, inf_cols = [], [], [], []
    try:
        dat = read_ascii(path)
    except:
        etype, emsg, etb = sys.exc_info()
        return nanresult(False, emsg, nan_rows, nan_cols, inf_rows, inf_cols)

    if np.all(np.isfinite(dat.data)):
        return nanresult(True, 'file ok', nan_rows, nan_cols, inf_rows, inf_cols)

    msg = 'unknown'
    nanvals = np.where(np.isnan(dat.data))
    if len(nanvals[0]) > 0:
        msg = 'has nans'
        for icol in nanvals[0]:
            if icol not in nan_cols:
                nan_cols.append(icol)
        for irow in nanvals[1]:
            if irow not in nan_rows:
                nan_rows.append(irow)

    infvals = np.where(np.isinf(dat.data))
    if len(infvals[0]) > 0:
        if len(msg) == 0:
            msg = 'has infs'
        else:
            msg = 'has nans and infs'
        for icol in infvals[0]:
            if icol not in inf_cols:
                inf_cols.append(icol)
        for irow in infvals[1]:
            if irow not in inf_rows:
                inf_rows.append(irow)

    return nanresult(False, msg, nan_rows, nan_cols, inf_rows, inf_cols)


def getfloats(txt, allow_times=True):
    """convert a line of numbers into a list of floats,
    as for reading a file with columnar numerical data.

    Arguments
    ---------
      txt           str, line of text to parse
      allow_times   bool, whether to support time stamps [True]

    Returns
    -------
      list, each entry either a float or None

    Notes
    -----
      The `allow_times` will try to support common date-time strings
      using the dateutil module, returning a numerical value as the
      Unix timestamp, using
          time.mktime(dateutil.parser.parse(word).timetuple())
    """
    words = [w.strip() for w in txt.replace(',', ' ').split()]
    mktime = time.mktime
    for i, w in enumerate(words):
        val = None
        try:
            val = float(w)
        except ValueError:
            try:
                val = mktime(dateparse(w).timetuple())
            except ValueError:
                pass
        words[i] = val
    return words

def colname(txt):
    return fix_varname(txt.strip().lower()).replace('.', '_')


def lformat(val, length=12):
    """Format a number with fixed-length format, somewhat like '%g' except that

        a) the length of the output string will be the requested length.
        b) positive numbers will have a leading blank.
        b) the precision will be as high as possible.
        c) trailing zeros will not be trimmed.

    The precision will typically be length-7, but may be better than
    that for values with absolute value between 1.e-5 and 1.e8.

    Arguments
    ---------
    val       value to be formatted
    length    length of output string

    Returns
    -------
    string of specified length.

    Notes
    ------
     Positive values will have leading blank.

    """
    try:
        expon = int(log10(abs(val)))
    except (OverflowError, ValueError):
        expon = 0
    length = max(length, 7)
    form = 'e'
    prec = length - 7
    if abs(expon) > 99:
        prec -= 1
    elif ((expon > 0 and expon < (prec+4)) or
          (expon <= 0 and -expon < (prec-1))):
        form = 'f'
        prec += 4
        if expon > 0:
            prec -= expon
    fmt = '{0: %i.%i%s}' % (length, prec, form)
    return fmt.format(val)

def parse_labelline(labelline, header):
    """
    parse the 'label line' for an ASCII file.


    This is meant to handle some special cases of XAFS data collected at a variety of sources
    """

    pass


def read_ascii(filename, labels=None, simple_labels=False,
               sort=False, sort_column=0, delimeter=None):
    """read a column ascii column file, returning a group containing the data
    extracted from the file.

    read_ascii(filename, labels=None, simple_labels=False, sort=False, sort_column=0)

    Arguments
    ---------
     filename (str)        name of file to read
     labels (list or None) list of labels to use for array labels [None]
     simple_labels (bool)  whether to force simple column labels (note 1) [False]
     delimeter (str)       string to use to split label line
     sort (bool)           whether to sort row data (note 2) [False]
     sort_column (int)     column to use for sorting (note 2) [0]

    Returns
    --------
      group containing data read from file

    Notes
    -----
      1. array labels.  If `labels` is `None` (he default value), column labels
         (and so, names of 1d arrays) will be tried to be determined from the
         file header.  This often means parsing the final header line, but
         tagged column files from several XAFS beamlines will be tried and used
         if matching.  Column labels may be like 'col1', 'col2', etc if suitable
         column labels cannot be guessed.
         These labels will be used as names for the 1-d arrays from each column.
         If `simple_labels` is  `True`, the names 'col1', 'col2' etc will be used
         regardless of the column labels found in the file.

      2. sorting.  Data can be sorted to be in increasing order of any column,
         by giving the column index (starting from 0).

      3. header parsing. If header lines are of the forms of
            KEY : VAL
            KEY = VAL
         these will be parsed into a 'attrs' dictionary in the returned group.


    The returned group will have a number of members:

       GROUP.filename: text name of the file
       GROUP.array_labels: array labels, names of 1-D arrays
       GROUP.data:     2-dimensional data (ncolumns, nrows)
       GROUP.header:   array of text lines of the header.
       GROUP.footer:   array of text lines of the footer (text after the block of numerical data)
       GROUP.attrs :   group of attributes parsed from header lines
    """
    if not os.path.isfile(filename):
        raise OSError("File not found: '%s'" % filename)
    if os.stat(filename).st_size > MAX_FILESIZE:
        raise OSError("File '%s' too big for read_ascii()" % filename)

    with open(filename, 'r') as fh:
        text = fh.read()

    text = text.replace('\r\n', '\n').replace('\r', '\n').split('\n')

    ncol = None
    data, footers, headers = [], [], []

    text.reverse()
    section = 'FOOTER'

    for line in text:
        line = line.strip()
        if len(line) < 1:
            continue
        # look for section transitions (going from bottom to top)
        if section == 'FOOTER' and not None in getfloats(line):
            section = 'DATA'
        elif section == 'DATA' and None in getfloats(line):
            section = 'HEADER'

        # act of current section:
        if section == 'FOOTER':
            footers.append(line)
        elif section == 'HEADER':
            headers.append(line)
        elif section == 'DATA':
            rowdat  = getfloats(line)
            if ncol is None:
                ncol = len(rowdat)
            elif ncol > len(rowdat):
                rowdat.extend([np.nan]*(ncol-len(rowdat)))
            elif ncol < len(rowdat):
                for i in data:
                    i.extend([np.nan]*(len(rowdat)-ncol))
                ncol = len(rowdat)
            data.append(rowdat)

    # reverse header, footer, data, convert to arrays
    footers.reverse()
    headers.reverse()
    data.reverse()
    data = np.array(data).transpose()

    # try to parse attributes from header text
    header_attrs = {}
    for hline in headers:
        hline = hline.strip().replace('\t', ' ')
        if len(hline) < 1: continue
        if hline[0] in COMMENTCHARS:
            hline = hline[1:].strip()
        keywds = []
        if ':' in hline: # keywords in  'x: 22'
            words = hline.split(':', 1)
            keywds = words[0].split()
        elif '=' in hline: # keywords in  'x = 22'
            words = hline.split('=', 1)
            keywds = words[0].split()
        if len(keywds) == 1:
            key = colname(keywds[0])
            if key.startswith('_'):
                key = key[1:]
            if len(words) > 1:
                header_attrs[key] = words[1].strip()

    if sort and sort_column >= 0 and sort_column < ncol:
         data = data[:, np.argsort(data[sort_column])]

    path, fname = os.path.split(filename)
    attrs = {'filename': filename}
    group = Group(name='ascii_file %s' % filename,
                  path=filename,
                  filename=fname,
                  header=headers,
                  data=data)

    if len(footers) > 0:
        group.footer = footers

    group.attrs = Group(name='header attributes from %s' % filename)
    for key, val in header_attrs.items():
        setattr(group.attrs, key, val)

    if isinstance(labels, str):
        for bchar in ',#@%|:*':
            labels = labels.replace(bchar, '')
        labels = labels.split()
    if labels is None and not simple_labels:
        bldat = guess_beamline(headers)(headers)
        labels = bldat.get_array_labels()
        if getattr(bldat, 'energy_units', 'eV') != 'eV':
            group.energy_units = bldat.energy_units
        if getattr(bldat, 'energy_column', 1) != 1:
            group.energy_column = bldat.energy_column
        if getattr(bldat, 'mono_dspace', -1) > 0:
            group.mono_dspace = bldat.mono_dspace

    set_array_labels(group, labels=labels, simple_labels=simple_labels)
    return group

def set_array_labels(group, labels=None, simple_labels=False,
                     save_oldarrays=False, _larch=None):

    """set array names for a group from its 2D `data` array.

    Arguments
    ----------
      labels (list of strings or None)  array of labels to use
      simple_labels (bool):   flag to use ('col1', 'col2', ...) [False]
      save_oldarrays (bool):  flag to save old array names [False]


    Returns
    -------
       group with newly named attributes of 1D array data, and
       an updated `array_labels` giving the mapping of `data`
       columns to attribute names.

    Notes
    ------
      1. if `simple_labels=True` it will overwrite any values in `labels`

      3. Array labels must be valid python names. If not enough labels
         are specified, or if name clashes arise, the array names may be
         modified, often by appending an underscore and letter or by using
         ('col1', 'col2', ...) etc.

      4. When `save_oldarrays` is `False` (the default), arrays named in the
         current `group.array_labels` will be erased.  Other arrays and
         attributes will not be changed.

    """
    write = sys.stdout.write
    if _larch is not None:
        write = _larch.writer.write
    if not hasattr(group, 'data'):
        write("cannot set array labels for group '%s': no `data`\n" % repr(group))
        return

    # clear old arrays, if desired
    oldlabels = getattr(group, 'array_labels', None)
    if oldlabels is not None and not save_oldarrays:
        for attr in oldlabels:
            if hasattr(group, attr):
                delattr(group, attr)

    ncols, nrow = group.data.shape

    ####
    # step 1: determine user-defined labels from input options
    # generating array `tlabels` for test labels
    #
    # generate simple column labels, used as backup
    clabels = ['col%i' % (i+1) for i in range(ncols)]

    if isinstance(labels, str):
        labels = labels.split()


    tlabels = labels
    # if simple column names requested or above failed, use simple column names
    if simple_labels or tlabels is None:
        tlabels = clabels

    ####
    # step 2: check input and correct problems
    # 2.a: check for not enough and too many labels
    if len(tlabels) < ncols:
        for i in range(len(tlabels), ncols):
            tlabels.append("col%i" % (i+1))
    elif len(tlabels) > ncols:
        tlabels = tlabels[:ncols]

    # 2.b: check for names that clash with group attributes
    # or that are repeated, append letter.
    reserved_names = ('data', 'array_labels', 'filename',
                      'attrs', 'header', 'footer')
    extras = string.ascii_lowercase
    labels = []
    for i in range(ncols):
        lname = tlabels[i]
        if lname in reserved_names or lname in labels:
            lname = lname + '_a'
            j = 0
            while lname in labels:
                j += 1
                if j == len(extras):
                    break
                lname = "%s_%s" % (tlabels[i], extras[j])
        if lname in labels:
            lname = clabels[i]
        labels.append(lname)

    ####
    # step 3: assign attribue names, set 'array_labels'
    for i, name in enumerate(labels):
        setattr(group, name, group.data[i])
    group.array_labels = labels
    return group


def write_ascii(filename, *args, commentchar='#', label=None, header=None,
                _larch=None, **kws):
    """write a list of items to an ASCII column file

    write_ascii(filename, arg1, arg2, arg3, ... **args)

    arguments
    ---------
    commentchar: character for comment ('#')
    label:       array label line (autogenerated)
    header:      array of strings for header

    """
    ARRAY_MINLEN = 2
    write = sys.stdout.write
    if _larch is not None:
        write = _larch.writer.write

    com = commentchar
    label = label
    if header is None:
        header = []

    arrays = []
    arraylen = None

    for arg in args:
        if isinstance(arg, np.ndarray):
            if len(arg) > ARRAY_MINLEN:
                if arraylen is None:
                    arraylen = len(arg)
                else:
                    arraylen = min(arraylen, len(arg))
                arrays.append(arg)
            else:
                header.append(repr(arg))

        else:
            header.append(repr(arg))

    if arraylen is None:
        raise ValueError("write_ascii() need %i or more elements in arrays." % ARRAY_MINLEN)

    buff = []
    if header is None:
        buff = ['%s Output from Larch %s' % (com, time.ctime())]
    for s in header:
        buff.append('%s %s' % (com, s))
    buff.append('%s---------------------------------'% com)
    if label is None:
        label = (' '*13).join(['col%i' % (i+1) for i in range(len(arrays))])
    buff.append('#  %s' % label)

    arrays = np.array(arrays)
    for i in range(arraylen):
        w = [" %s" % lformat(val[i], length=14) for val in arrays]
        buff.append('  '.join(w))

    try:
        fout = open(filename, 'w')
    except:
        write("cannot open file %s'\n" % filename)
        return

    try:
        fout.write('\n'.join(buff))
        fout.write('\n')
    except:
        write("cannot write to file %s'\n" % filename)
        return
    write("wrote to file '%s'\n" % filename)


def write_group(filename, group, scalars=None, arrays=None,
                arrays_like=None, commentchar='#', _larch=None):
    """write components of a group to an ASCII column file

    write_group(filename, group, commentchar='#')

    Warning: This is pretty minimal and may work poorly
    for large groups of complex data.
    """

    items = dir(group)
    npts = 0
    if arrays is None:
        arrays = []
    if scalars is None:
        scalars = []

    if arrays_like is not None and arrays_like in items:
        array = getattr(group, arrays_like)
        if isinstance(array, np.ndarray):
            npts = len(array)

    for name in items:
        val = getattr(group, name)
        if isinstance(val, np.ndarray):
            if npts != 0 and npts == len(val) and name not in arrays:
                arrays.append(name)

    header =[]
    for s in scalars:
        if s in items:
            val = getattr(group, s)
            header.append("%s = %s" % (s, val))

    label = '  '.join(arrays)

    args = []
    for name in arrays:
        if name in items:
            args.append(getattr(group, name))

    write_ascii(filename, *args, commentchar=commentchar,
                label=label, header=header, _larch=_larch)


def guess_filereader(filename):
    """guess function name to use to read an ASCII data file based
    on the file header

    Arguments
    ---------
    filename (str)   name of file to be read

    Returns
    -------
      name of function (as a string) to use to read file
    """
    with open(path, 'r') as fh:
        line1 = fh.readline()
    line1 = lines[0].lower()

    reader = 'read_ascii'
    if 'xdi' in line1:
        reader = 'read_xdi'
        if ('epics stepscan' in line1 or 'gse' in line1):
            reader = 'read_gsexdi'
    elif 'epics scan' in line1:
        reader = 'read_gsescan'
    return reader



#!/usr/bin/env python
"""
XRF Analysis Panel
"""
import os
import json
from functools import partial

import wx
import wx.lib.scrolledpanel as scrolled

import numpy as np

from ..wxlib import (LarchPanel, LarchFrame, EditableListBox, SimpleText,
                     FloatCtrl, Font, pack, Popup, Button, MenuItem,
                     Choice, Check, GridPanel, FileSave, FileOpen, HLine)
from ..utils.strutils import bytes2str, version_ge, fix_varname

from ..xrmmap import GSEXRM_MapFile, GSEXRM_FileStatus, h5str, ensure_subgroup

from ..wxlib.xrfdisplay_utils import (XRFGROUP, mcaname,
                                      XRFRESULTS_GROUP,
                                      MAKE_XRFRESULTS_GROUP)

CEN = wx.ALIGN_CENTER
LEFT = wx.ALIGN_LEFT
RIGHT = wx.ALIGN_RIGHT
ALL_CEN =  wx.ALL|CEN
ALL_LEFT =  wx.ALL|LEFT

NOFITS_MSG  = "No XRF Fits: Use XRFViewer to fit spectrum."
HASFITS_MSG = "Select XRF Fit to build elemental maps"

from ..wxlib.xrfdisplay_utils import (XRFGROUP, MAKE_XRFGROUP_CMD, next_mcaname)

class XRFAnalysisPanel(scrolled.ScrolledPanel):
    """Panel of XRF Analysis results"""
    label  = 'XRF Analysis'
    def __init__(self, parent, owner=None, **kws):
        self.owner = owner
        self.map   = None
        self.cfile = None
        scrolled.ScrolledPanel.__init__(self, parent, -1,
                                        style=wx.GROW|wx.TAB_TRAVERSAL, **kws)
        sizer = wx.GridBagSizer(3, 3)
        self.current_fit = 0
        self.fit_choice = Choice(self, size=(375, -1), choices=[])

        self.use_nnls = Check(self, label='force non-negative (~5x slower)',
                              default=False)

        save_btn = Button(self, 'Calculate Element Maps', size=(200, -1),
                          action=self.onSaveArrays)

        load_btn = Button(self, 'Load Saved XRF Model', size=(200, -1),
                          action=self.onLoadXRFModel)

        self.scale = FloatCtrl(self, value=1.0, minval=0, precision=5, size=(100, -1))
        self.name  = wx.TextCtrl(self, value='abundance', size=(375, -1))

        self.fit_status = SimpleText(self, label=NOFITS_MSG)

        ir = 0
        sizer.Add(self.fit_status,           (ir, 0), (1, 2), ALL_LEFT, 2)
        sizer.Add(load_btn,                  (ir, 2), (1, 2), ALL_LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(self, 'Use Fit Label:'),  (ir, 0), (1, 1), ALL_LEFT, 2)
        sizer.Add(self.fit_choice,                     (ir, 1), (1, 3), ALL_LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(self, 'Scaling Factor:'), (ir, 0), (1, 1), ALL_LEFT, 2)
        sizer.Add(self.scale,                          (ir, 1), (1, 1), ALL_LEFT, 2)
        sizer.Add(self.use_nnls,                       (ir, 2), (1, 2), ALL_LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(self, 'Save to Group:'), (ir, 0), (1, 1), ALL_LEFT, 2)
        sizer.Add(self.name,                          (ir, 1), (1, 3), ALL_LEFT, 2)
        ir += 1
        sizer.Add(save_btn,                           (ir, 0), (1, 3), ALL_LEFT, 2)

        ir += 1
        sizer.Add(HLine(self, size=(600, 5)), (ir, 0), (1, 4), ALL_LEFT, 2)

        pack(self, sizer)
        self.SetupScrolling()

    def onSaveArrays(self, evt=None):
        self.current_fit = cfit = self.fit_choice.GetSelection()
        try:
            thisfit = self.owner.larch.symtable._xrfresults[cfit]
        except:
            thisfit = None
        if thisfit is None:
            print("Unknown fit? " , cfit,
                  self.owner.larch.symtable._xrfresults)
            return

        scale = self.scale.GetValue()
        method = 'nnls' if self.use_nnls.IsChecked() else 'lstsq'
        xrmfile = self.owner.current_file
        workname = fix_varname(self.name.GetValue())

        cmd = """weights = _xrfresults[{cfit:d}].decompose_map({groupname:s}.xrmmap['mcasum/counts'],
        scale={scale:.6f}, pixel_time={ptime:.5f},method='{method:s}')
        for key, val in weights.items():
             {groupname:s}.add_work_array(val, fix_varname(key), parent='{workname:s}')
        #endfor
        """
        cmd = cmd.format(cfit=cfit, groupname=xrmfile.groupname, ptime=xrmfile.pixeltime,
                         workname=workname, scale=scale, method=method)
        self.owner.larch.eval(cmd)
        dlist = xrmfile.get_detector_list(use_cache=False)
        for p in self.owner.nb.pagelist:
            if hasattr(p, 'update_xrmmap'):
                p.detectors_set = False
                p.update_xrmmap(xrmfile=xrmfile, set_detectors=True)

    def onLoadXRFModel(self, evt=None):
        _larch = self.owner.larch
        symtab = _larch.symtable
        FILE_WILDCARDS = "XRF Model Files(*.xrfmodel)|*.xrfmodel|All files (*.*)|*.*"

        dlg = wx.FileDialog(self, message="Read XRF Model File",
                            defaultDir=os.getcwd(),
                            wildcard=FILE_WILDCARDS,
                            style=wx.FD_OPEN)
        path = None
        if dlg.ShowModal() == wx.ID_OK:
            path = dlg.GetPath()
        dlg.Destroy()
        if path is not None:
            _, fname = os.path.split(path)
            if not symtab.has_group(XRFRESULTS_GROUP):
                _larch.eval(MAKE_XRFRESULTS_GROUP)
            _larch.eval("tmp = xrf_fitresult('{:s}')".format(path))
            _larch.eval("tmp.filename = '{:s}'".format(fname))
            _larch.eval("_xrfresults.insert(0, tmp)")
            _larch.eval("del tmp")
            self.current_fit = 0
            self.update_xrmmap()

    def onOtherSaveArray(self, evt=None):
        name = self.name_in.GetValue()
        expr = self.expr_in.GetValue()
        xrmfile = self.owner.current_file
        info = []
        for varname in sorted(self.varfile.keys()):
            fname   = self.varfile[varname].GetStringSelection()
            roiname = self.varroi[varname].GetStringSelection()
            dname   = self.vardet[varname].GetStringSelection()
            dtcorr  = self.varcor[varname].IsChecked()
            info.append((varname, (fname, roiname, dname, dtcorr)))

        if self.map is None:
            self.onShowMap()

        if name in xrmfile.work_array_names():
            if (wx.ID_YES == Popup(self.owner, """Overwrite Array '%s' for %s?
    WARNING: This cannot be undone
    """ % (name, xrmfile.filename),
                                   'Overwrite Array?', style=wx.YES_NO)):
                xrmfile.del_work_array(h5str(name))
            else:
                return
        xrmfile.add_work_array(self.map, h5str(name),
                               expression=h5str(expr),
                               info=json.dumps(info))

        for p in self.owner.nb.pagelist:
            if hasattr(p, 'update_xrmmap'):
                p.update_xrmmap(xrmfile=xrmfile, set_detectors=True)

    def update_xrmmap(self, xrmfile=None, set_detectors=None):
        if xrmfile is None:
            xrmfile = self.owner.current_file
        self.cfile = xrmfile
        symtab = self.owner.larch.symtable
        xrfresults = getattr(symtab, '_xrfresults', [])
        fit_names = ["%s: %s" % (a.label, a.mcalabel) for a in xrfresults]

        if len(fit_names) > 0:
            self.fit_status.SetLabel(HASFITS_MSG)
            self.fit_choice.Clear()
            self.fit_choice.AppendItems(fit_names)
            self.fit_choice.SetSelection(self.current_fit)



#!/usr/bin/env python
# coding: utf-8
"""
RIXS data model
"""
from silx.gui import qt
from larch.qtlib.model import TreeModel


class RixsTreeModel(TreeModel):
    """TreeModel customized for RIXS"""

    def __init__(self, parent=None):

        super(RixsTreeModel, self).__init__(parent=parent)


class RixsListModel(qt.QAbstractListModel):
    """Simple ListModel used in larch.qtrixs.plotrixs.RixsMainWindow"""

    def __init__(self, *args, data=None, **kwargs):
        """Constructor"""

        super(RixsListModel, self).__init__(*args, **kwargs)
        self._data = data or []

    def data(self, index, role):
        if role == qt.Qt.DisplayRole:
            row = index.row()
            rd = self._data[row]
            _name = "{0}: {1}".format(row, rd.sample_name)
            return _name

    def rowCount(self, index):
        return len(self._data)

    def appendRow(self, rxobj):
        self._data.append(rxobj)
        self.layoutChanged.emit()


if __name__ == '__main__':
    pass




import time
import numpy as np
from scipy.special import erfc

from larch import Group
from larch.math import interp
from larch.utils.show import show

from .mback import mback

try:
    from larch.wxlib   import _newplot, _plot
    HASPLOT = True
except:
    HASPLOT = False

pi = np.pi

TINY = 1e-20
FOPI = 4/pi
MAXORDER = 6

##
## to test the advantage of the vectorized MacLaurin series algorithm:
##
#     import time
#     a=read_ascii('/home/bruce/git/demeter/examples/cu/cu10k.dat')
#
#     b=diffkk(a.energy, a.xmu, e0=8979, z=29, order=4, form='mback')
#     start = time.clock()
#     for i in range(10):
#         b.kktrans()
#     endfor
#     finish = time.clock()
#     print(finish - start)
#
#     start = time.clock()
#     for i in range(10):
#         b.kktrans(how='sca')
#     endfor
#     finish = time.clock()
#     print(finish - start)
##
## I got 7.6 seconds and 76.1 seconds for 10 iterations of each.  Awesome!
##


###
###  These are the scalar forms of the MacLaurin series algorithm as originally coded up by Matt
###  see https://github.com/newville/ifeffit/blob/master/src/diffkk/kkmclr.f,
###      https://github.com/newville/ifeffit/blob/master/src/diffkk/kkmclf.f,
###  and https://gist.github.com/maurov/33997083e96ab4036fe7
###  The last one is a direct translation of Fortan to python, written by Matt, and used in
###  CARD (http://www.esrf.eu/computing/scientific/CARD/CARD.html)
###
###  See below for vector forms (much faster!)
###
def kkmclf_sca(e, finp):
    """
    forward (f'->f'') kk transform, using mclaurin series algorithm

    arguments:
    npts   size of arrays to consider
    e      energy array *must be on an even grid with an even number of points* [npts] (in)
    finp   f' array [npts] (in)
    fout   f'' array [npts] (out)
    notes  fopi = 4/pi
    """
    npts = len(e)
    if npts != len(finp):
        Exception("Input arrays not of same length in kkmclf_sca")
    if npts < 2:
        Exception("Array too short in kkmclf_sca")
    if npts % 2:
        Exception("Array has an odd number of elements in kkmclf_sca")
    fout = [0.0]*npts
    if npts >= 2:
        factor = FOPI * (e[npts-1] - e[0]) / (npts - 1)
        nptsk = npts / 2
        for i in range(npts):
            fout[i] = 0.0
            ei2 = e[i]*e[i]
            ioff = i%2 - 1
            for k in range(nptsk):
                j = k + k + ioff
                de2 = e[j]*e[j] - ei2
                if abs(de2) <= TINY:
                    de2 = TINY
                fout[i] = fout[i] + finp[j]/de2
            fout[i] *= factor*e[i]
    return fout

def kkmclr_sca(e, finp):
    """
    reverse (f''->f') kk transform, using maclaurin series algorithm

    arguments:
    npts   size of arrays to consider
    e      energy array *must be on a even grid with an even number of points* [npts] (in)
    finp   f'' array [npts] (in)
    fout   f' array [npts] (out)
    m newville jan 1997
    """
    npts = len(e)
    if npts != len(finp):
        Exception("Input arrays not of same length in kkmclr")
    if npts < 2:
        Exception("Array too short in kkmclr")
    if npts % 2:
        Exception("Array has an odd number of elements in kkmclr")
    fout = [0.0]*npts

    factor = -FOPI * (e[npts-1] - e[0]) / (npts - 1)
    nptsk  = npts / 2
    for i in range(npts):
        fout[i] = 0.0
        ei2 = e[i]*e[i]
        ioff = i%2 - 1
        for k in range(nptsk):
            j = k + k + ioff
            de2 = e[j]*e[j] - ei2
            if abs(de2) <= TINY:
                de2 = TINY
            fout[i] = fout[i] + e[j]*finp[j]/de2
        fout[i] *= factor
    return fout

###
###  These are vector forms of the MacLaurin series algorithm, adapted from Matt's code by Bruce
###  They are about an order of magnitude faster.
###

def kkmclf(e, finp):
    """
    forward (f'->f'') kk transform, using maclaurin series algorithm

    arguments:
      e      energy array *must be on an even grid with an even number of points* [npts] (in)
      finp   f'' array [npts] (in)
      fout   f' array [npts] (out)
    """
    npts = len(e)
    if npts != len(finp):
        Exception("Input arrays not of same length for diff KK transform in kkmclr")
    if npts < 2:
        Exception("Array too short for diff KK transform in kkmclr")
    if npts % 2:
        Exception("Array has an odd number of elements for diff KK transform in kkmclr")

    fout   = np.zeros(npts)
    factor = FOPI * (e[-1] - e[0]) / (npts-1)
    ei2    = e**2
    ioff   = np.mod(np.arange(npts), 2) - 1

    nptsk  = int(npts/2)
    k      = np.arange(nptsk)

    for i in range(npts):
        j    = 2*k + ioff[i]
        de2  = e[j]**2 - ei2[i]
        fout[i] = sum(finp[j]/de2)

    fout = fout * factor
    return fout


def kkmclr(e, finp):
    """
    reverse (f''->f') kk transform, using maclaurin series algorithm

    arguments:
      e      energy array *must be on an even grid with an even number of points* [npts] (in)
      finp   f' array [npts] (in)
      fout   f'' array [npts] (out)
    """
    npts = len(e)
    if npts != len(finp):
        Exception("Input arrays not of same for length diff KK transform in kkmclr")
    if npts < 2:
        Exception("Array too short for diff KK transform in kkmclr")
    if npts % 2:
        Exception("Array has an odd number of elements for diff KK transform in kkmclr")

    fout   = np.zeros(npts)
    factor = -FOPI * (e[-1] - e[0]) / (npts-1)
    ei2    = e**2
    ioff   = np.mod(np.arange(npts), 2) - 1

    nptsk  = int(npts/2)
    k      = np.arange(nptsk)

    for i in range(npts):
        j    = 2*k + ioff[i]
        de2  = e[j]**2 - ei2[i]
        fout[i] = sum(e[j]*finp[j]/de2)

    fout = fout * factor
    return fout


class diffKKGroup(Group):
    """
    A Larch Group for generating f'(E) and f"(E) from a XAS measurement of mu(E).
    """

    def __init__(self, energy=None, mu=None, z=None, edge='K', mback_kws=None, **kws):
        kwargs = dict(name='diffKK')
        kwargs.update(kws)
        Group.__init__(self,  **kwargs)
        self.energy     = energy
        self.mu         = mu
        self.z          = z
        self.edge       = edge
        self.mback_kws  = mback_kws
    def __repr__(self):
        return '<diffKK Group>'


# e0=None, z=None, edge=None, order=3, form='mback', whiteline=False, how=None
    def kk(self, energy=None, mu=None, z=None, edge='K', how='scalar', mback_kws=None):
        """
        Convert mu(E) data into f'(E) and f"(E).  f"(E) is made by
        matching mu(E) to the tabulated values of the imaginary part
        of the scattering factor (Cromer-Liberman), f'(E) is then
        obtained by performing a differential Kramers-Kronig transform
        on the matched f"(E).

          Attributes
            energy:     energy array
            mu:         array with mu(E) data
            z:          Z number of absorber
            edge:       absorption edge, usually 'K' or 'L3'
            mback_kws:  arguments for the mback algorithm

          Returns
            self.f1, self.f2:  CL values over on the input energy grid
            self.fp, self.fpp: matched and KK transformed data on the input energy grid

        References:
          * Cromer-Liberman: http://dx.doi.org/10.1063/1.1674266
          * KK computation: Ohta and Ishida, Applied Spectroscopy 42:6 (1988) 952-957
          * diffKK implementation: http://dx.doi.org/10.1103/PhysRevB.58.11215
          * MBACK (Weng, Waldo, Penner-Hahn): http://dx.doi.org/10.1086/303711
          * Lee and Xiang: http://dx.doi.org/10.1088/0004-637X/702/2/970

        """

        if type(energy).__name__ == 'ndarray': self.energy = energy
        if type(mu).__name__     == 'ndarray': self.mu     = mu
        if z    != None: self.z    = z
        if edge != None: self.edge = edge
        if mback_kws != None: self.mback_kws = mback_kws

        if self.z == None:
            Exception("Z for absorber not provided for diffKK")
        if self.edge == None:
            Exception("absorption edge not provided for diffKK")

        mb_kws = dict(order=3, z=self.z, edge=self.edge, e0=None,
                      leexiang=False, tables='chantler',
                      fit_erfc=False, return_f1=True)
        if self.mback_kws is not None:
            mb_kws.update(self.mback_kws)

        start = time.monotonic()

        mback(self.energy, self.mu, group=self, **mb_kws)

        ## interpolate matched data onto an even grid with an even number of elements (about 1 eV)
        npts = int(self.energy[-1] - self.energy[0]) + (int(self.energy[-1] - self.energy[0])%2)
        self.grid = np.linspace(self.energy[0], self.energy[-1], npts)
        fpp = interp(self.energy, self.f2-self.fpp, self.grid, fill_value=0.0)

        ## do difference KK
        if repr(how).startswith('sca'):
            fp = kkmclr_sca(self.grid, fpp)
        else:
            fp = kkmclr(self.grid, fpp)

        ## interpolate back to original grid and add diffKK result to f1 to make fp array
        self.fp = self.f1 + interp(self.grid, fp, self.energy, fill_value=0.0)

        ## clean up group
        #for att in ('normalization_function', 'weight', 'grid'):
        #    if hasattr(self, att): delattr(self, att)
        finish = time.monotonic()
        self.time_elapsed = float(finish-start)

def diffkk(energy=None, mu=None, z=None, edge='K', mback_kws=None, **kws):
    """
    Make a diffKK group given mu(E) data

      Attributes
        energy:     energy array
        mu:         array with mu(E) data
        z:          Z number of absorber
        edge:       absorption edge, usually 'K' or 'L3'
        mback_kws:  arguments for the mback algorithm
    """
    return diffKKGroup(energy=energy, mu=mu, z=z, mback_kws=mback_kws)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Application window of RIXS GUI
==============================
"""
from silx.gui import qt

from larch.utils.logging import getLogger
from larch.qtlib.console import InternalIPyKernel
from larch.qtlib.model import HeaderSection
from larch.qtlib.delegates import ComboBoxDelegate

from .plotrixs import RixsPlotArea
from .view import RixsTreeView as RixsView
from .model import RixsTreeModel as RixsModel
from .items import RixsItem


class RixsAppWindow(qt.QMainWindow):
    """MainWindow may also behave as widget"""

    def __init__(self, parent=None, with_ipykernel=True, logger=None):
        """Constructor"""

        self._logger = logger or getLogger('RixsAppWindow')

        super(RixsAppWindow, self).__init__(parent=parent)

        if parent is not None:
            #: behave as a widget
            self.setWindowFlags(qt.Qt.Widget)
        else:
            #: main window
            self.setWindowTitle('RIXS_VIEW')

            # TODO: Add icon to the application
            #ico = qt.QIcon(os.path.join(_resourcesPath, "logo",
            #                            "xraysloth_logo_04.svg"))
            #self.setWindowIcon(ico)

        #: IPython kernel status
        self._with_ipykernel = with_ipykernel

        #: Model/view
        self._model = RixsModel()
        self._view = RixsView(parent=self)
        self._view.setModel(self._model)

        # Add additional sections to the header.
        values = [
            HeaderSection(name='Plot',
                          roles={qt.Qt.DisplayRole: 'currentPlotWindowIndex',
                                 qt.Qt.EditRole: 'plotWindowsIndexes'
                                 },
                          delegate=ComboBoxDelegate),
            ]

        for value in values:
            section = len(self._model.header)
            orientation = qt.Qt.Horizontal
            self._model.setHeaderData(section, orientation, value)

        # Add (empty) menu bar -> contents added later
        self._menuBar = qt.QMenuBar()
        self.setMenuBar(self._menuBar)
        self._initAppMenu()

        #: Plot Area
        self._plotArea = RixsPlotArea(self)
        self.setCentralWidget(self._plotArea)

        #: TreeView dock widget
        self._dockDataWidget = qt.QDockWidget(parent=self)
        self._dockDataWidget.setObjectName('Data View')
        self._dockDataWidget.setWidget(self._view)
        self.addDockWidget(qt.Qt.LeftDockWidgetArea, self._dockDataWidget)

        #: Plots update
        self._model.dataChanged.connect(self.updatePlot)
        self._plotArea.changed.connect(self.updateModel)

        #: Console
        if self._with_ipykernel:
            # Initialize internal ipykernel
            self._ipykernel = InternalIPyKernel()
            self._ipykernel.init_kernel(backend='qt')
            self._ipykernel.add_to_namespace('view', self._view)
            self._ipykernel.add_to_namespace('model', self._model)
            self._ipykernel.add_to_namespace('plot', self._plotArea)

            # Add IPython console at menu
            self._initConsoleMenu()
        else:
            self._ipykernel = None

    def updateModel(self):
        plotWindows = self._plotArea.plotWindows()
        for item in self._view.rixsItems():
            item.plotWindows = plotWindows
            if len(plotWindows) == 0:
                index = self._model.indexFromItem(item)
                self._model.dataChanged.emit(index, index)

    def updatePlot(self, *args):
        topLeft, bottomRight, _ = args

        topLeftItem = self._model.itemFromIndex(topLeft)
        bottomRightItem = self._model.itemFromIndex(bottomRight)

        if topLeftItem is not bottomRightItem:
            self._logger.error('The indices do not point to the same item in the model')
            return

        item = topLeftItem
        plotWindows = self._plotArea.plotWindows()

        if item.isChecked:
            if len(plotWindows) == 0:
                self._logger.info('There are no plot widgets available')
                return

        for plotWindow in plotWindows:
            plotWindow.remove(item.legend)
            if not plotWindow.getItems():
                plotWindow.reset()
            else:
                plotWindow.statusBar().clearMessage()

        rixsItems = self._view.rixsItems()
        if item.isChecked:
            if item in list(rixsItems) and isinstance(item, RixsItem):
                item.plot()

    def showEvent(self, event):
        self.loadSettings()
        super(RixsAppWindow, self).showEvent(event)

    def closeEvent(self, event):
        self.saveSettings()
        super(RixsAppWindow, self).closeEvent(event)

    def loadSettings(self):
        """TODO"""
        pass

    def saveSettings(self):
        """TODO"""
        pass

    # Populate the menu bar with common actions and shortcuts
    def _addMenuAction(self, menu, action, deferShortcut=False):
        """Add action to menu as well as self so that when the menu bar is
        invisible, its actions are still available. If deferShortcut
        is True, set the shortcut context to widget-only, where it
        will avoid conflict with shortcuts already bound to the
        widgets themselves.
        """
        menu.addAction(action)
        self.addAction(action)

        if deferShortcut:
            action.setShortcutContext(qt.Qt.WidgetShortcut)
        else:
            action.setShortcutContext(qt.Qt.ApplicationShortcut)

    def _initAppMenu(self):
        """Add application menu"""
        self._menuApp = self._menuBar.addMenu("Application")
        self._closeAppAction = qt.QAction("&Quit", self, shortcut="Ctrl+Q",
                                          triggered=self.onClose)
        self._addMenuAction(self._menuApp, self._closeAppAction)

    def _initConsoleMenu(self):
        self._menuConsole = self._menuBar.addMenu("Console")

        self._newConsoleAction = qt.QAction("&New Qt Console",
                                            self, shortcut="Ctrl+K",
                                            triggered=self._ipykernel.new_qt_console)
        self._addMenuAction(self._menuConsole, self._newConsoleAction)

    def onClose(self):
        if self._ipykernel is not None:
            self._ipykernel.cleanup_consoles()
        self.closeEvent(quit())


def main():
    from silx import sx
    sx.enable_gui()
    rx = RixsAppWindow()
    rx.show()
    input("Press ENTER to close the window...")


if __name__ == '__main__':
    main()
